/*
 *  C64.cpp - Put the pieces together
 *
 *  Frodo (C) 1994-1997,2002 Christian Bauer
 */

#include "sysdeps.h"

#include "C64.h"
#include "CPUC64.h"
#include "CPU1541.h"
#include "VIC.h"
#include "SID.h"
#include "CIA.h"
#include "IEC.h"
#include "1541job.h"
#include "Display.h"
#include "Prefs.h"


/*
 *  Constructor: Allocate objects and memory
 */

C64::C64(ArmletData *d)
{
	int i,j;
	uint8 *p;

	armletData=d;

	// The thread is not yet running
	thread_running = false;
	quit_thyself = false;
	have_a_break = false;

	// System-dependent things
	c64_ctor1();

	// Open display
	TheDisplay = new C64Display(this);
	if (!TheDisplay) 
	{
		if (!d->errCode) 
			d->errCode=ERR_NEW_DISPLAY;
		return;
	}

	// Create the chips
	TheCPU = new MOS6510(this, RAM, Basic, Kernal, Char, Color, IO_Ram);
	if (!TheCPU) 
	{
		if (!d->errCode) 
			d->errCode=ERR_NEW_CPU;
		return;
	}

	TheJob1541 = new Job1541(this, RAM1541);
	if (!TheJob1541) 
	{
		if (!d->errCode) 
			d->errCode=ERR_NEW_JOB1541;
		return;
	}

	TheCPU1541 = new MOS6502_1541(this, TheJob1541, TheDisplay, RAM1541, ROM1541);
	if (!TheCPU1541) 
	{
		if (!d->errCode) 
			d->errCode=ERR_NEW_CPU1541;
		return;
	}

	TheVIC = TheCPU->TheVIC = new MOS6569(this, TheDisplay, TheCPU, RAM, Char, Color);
	if (!TheVIC) 
	{
		if (!d->errCode) 
			d->errCode=ERR_NEW_VIC;
		return;
	}

	TheSID = TheCPU->TheSID = new MOS6581(this);
	if (!TheSID) 
	{
		if (!d->errCode) 
			d->errCode=ERR_NEW_SID;
		return;
	}

	TheCIA1 = TheCPU->TheCIA1 = new MOS6526_1(TheCPU, TheVIC);
	if (!TheCIA1) 
	{
		if (!d->errCode) 
			d->errCode=ERR_NEW_CIA1;
		return;
	}

	TheCIA2 = TheCPU->TheCIA2 = TheCPU1541->TheCIA2 = new MOS6526_2(TheCPU, TheVIC, TheCPU1541);
	if (!TheCIA2) 
	{
		if (!d->errCode) 
			d->errCode=ERR_NEW_CIA2;
		return;
	}

	TheIEC = TheCPU->TheIEC = new IEC(this, TheDisplay);
	if (!TheIEC) 
	{
		if (!d->errCode) 
			d->errCode=ERR_NEW_IEC;
		return;
	}

	// Initialize RAM with powerup pattern
	for (i=0, p=RAM; i<512; i++) 
	{
		for (j=0; j<64; j++)
			*p++ = 0;
		for (j=0; j<64; j++)
			*p++ = 0xff;
	}

	// Initialize color RAM with random values
	for (i=0, p=Color; i<1024; i++)
	{
		*p++ = Random() & 0x0f;
	}

	// Clear 1541 RAM
	memset(RAM1541, 0, 0x800);

	CycleCounter = 0;
	
	SwitchToSC = false;
	SwitchToStandard = false;
}


/*
 *  Destructor: Delete all objects
 */

C64::~C64()
{
	delete TheJob1541;
	delete TheIEC;
	delete TheCIA2;
	delete TheCIA1;
	delete TheSID;
	delete TheVIC;
	delete TheCPU1541;
	delete TheCPU;
	delete TheDisplay;
}


/*
 *  Reset C64
 */

void C64::Reset(void)
{
	TheCPU->AsyncReset();
	TheCPU1541->AsyncReset();
	TheSID->Reset();
	TheCIA1->Reset();
	TheCIA2->Reset();
	TheIEC->Reset();
}


/*
 *  NMI C64
 */

void C64::NMI(void)
{
	TheCPU->AsyncNMI();
}


/*
 *  The preferences have changed. prefs is a pointer to the new
 *   preferences, ThePrefs still holds the previous ones.
 *   The emulation must be in the paused state!
 */

void C64::NewPrefs(Prefs *prefs)
{
	PatchKernal(prefs->FastReset, prefs->Emul1541Proc);
	TheIEC->NewPrefs(prefs);
	TheJob1541->NewPrefs(prefs);
	TheSID->NewPrefs(prefs);

  if(!ThePrefs.SingleCycleEmulation && prefs->SingleCycleEmulation)
  {
    // Single cycle emulation switched on
    SwitchToSC = true;
  }
  
  if(ThePrefs.SingleCycleEmulation && !prefs->SingleCycleEmulation)
  {
    // Single cycle emulation switched off
    SwitchToStandard = true;
  }
  
	// Reset 1541 processor if turned on
	if (!ThePrefs.Emul1541Proc && prefs->Emul1541Proc)
	{
		TheCPU1541->AsyncReset();
	}
}


/*
 *  Patch kernal IEC routines
 */

void C64::PatchKernal(bool fast_reset, bool emul_1541_proc)
{
	if (fast_reset) 
	{
		Kernal[0x1d84] = 0xa0;
		Kernal[0x1d85] = 0x00;
	} 
	else 
	{
		Kernal[0x1d84] = orig_kernal_1d84;
		Kernal[0x1d85] = orig_kernal_1d85;
	}

	if (emul_1541_proc) 
	{
		Kernal[0x0d40] = 0x78;
		Kernal[0x0d41] = 0x20;
		Kernal[0x0d23] = 0x78;
		Kernal[0x0d24] = 0x20;
		Kernal[0x0d36] = 0x78;
		Kernal[0x0d37] = 0x20;
		Kernal[0x0e13] = 0x78;
		Kernal[0x0e14] = 0xa9;
		Kernal[0x0def] = 0x78;
		Kernal[0x0df0] = 0x20;
		Kernal[0x0dbe] = 0xad;
		Kernal[0x0dbf] = 0x00;
		Kernal[0x0dcc] = 0x78;
		Kernal[0x0dcd] = 0x20;
		Kernal[0x0e03] = 0x20;
		Kernal[0x0e04] = 0xbe;
	} 
	else 
	{
		Kernal[0x0d40] = 0xf2;	// IECOut
		Kernal[0x0d41] = 0x00;
		Kernal[0x0d23] = 0xf2;	// IECOutATN
		Kernal[0x0d24] = 0x01;
		Kernal[0x0d36] = 0xf2;	// IECOutSec
		Kernal[0x0d37] = 0x02;
		Kernal[0x0e13] = 0xf2;	// IECIn
		Kernal[0x0e14] = 0x03;
		Kernal[0x0def] = 0xf2;	// IECSetATN
		Kernal[0x0df0] = 0x04;
		Kernal[0x0dbe] = 0xf2;	// IECRelATN
		Kernal[0x0dbf] = 0x05;
		Kernal[0x0dcc] = 0xf2;	// IECTurnaround
		Kernal[0x0dcd] = 0x06;
		Kernal[0x0e03] = 0xf2;	// IECRelease
		Kernal[0x0e04] = 0x07;
	}

	// 1541
	ROM1541[0x2ae4] = 0xea;		// Don't check ROM checksum
	ROM1541[0x2ae5] = 0xea;
	ROM1541[0x2ae8] = 0xea;
	ROM1541[0x2ae9] = 0xea;
	ROM1541[0x2c9b] = 0xf2;		// DOS idle loop
	ROM1541[0x2c9c] = 0x00;
	ROM1541[0x3594] = 0x20;		// Write sector
	ROM1541[0x3595] = 0xf2;
	ROM1541[0x3596] = 0xf5;
	ROM1541[0x3597] = 0xf2;
	ROM1541[0x3598] = 0x01;
	ROM1541[0x3b0c] = 0xf2;		// Format track
	ROM1541[0x3b0d] = 0x02;
}


/*
 *  Save CPU state to snapshot
 *
 *  number of bytes written to p2 (p1 always 0x9000)
 */

uint32 C64::SaveCPUState(uint8 *p1, uint8 *p2)
{
	MOS6510State *state = (MOS6510State *) (((UInt32) p2+0x8000+0x403) & 0xfffffffc);
	TheCPU->GetState(state);

	memcpy(p1, RAM, 0x8000);
	memcpy(p1+0x8000, IO_Ram, 0x1000);
	memcpy(p2, RAM + 0x8000, 0x8000);
	memcpy(p2+0x8000, Color, 0x400);

	// p1 now contains 36k of memory
	// p2 contains other 32k, then Color and CPU state

	return ((UInt32) state + sizeof(MOS6510State) - (UInt32) p2);
}


/*
 *  Load CPU state from snapshot
 */

uint16 C64::LoadCPUState(uint8 *p1, uint8 *p2)
{
	MOS6510State *state = (MOS6510State *) (((UInt32) p2+0x8000+0x403) & 0xfffffffc);
	
	memcpy(RAM, p1, 0x8000);
	memcpy(IO_Ram, p1+0x8000, 0x1000);
	memcpy(RAM + 0x8000, p2, 0x8000);
	memcpy(Color, p2+0x8000, 0x400);
	TheCPU->SetState(state);

	return ((UInt32) state + sizeof(MOS6510State) - (UInt32) p2);
}

uint16 C64::LoadCPUStateOld(uint8 *p1, uint8 *p2)
{
	MOS6510State state;
	MOS6510StateOld *stateOld = (MOS6510StateOld *) (((UInt32) p2+0x8000+0x403) & 0xfffffffc);

	state.ar = 0;
	state.ar2 = 0;			// Address registers
	state.state = 0;
	state.op = 0;		// Current state and opcode
	state.rdbuf = 0;			// Data buffer for RMW instructions

	state.a = stateOld->a;
	state.x = stateOld->x;
	state.y = stateOld->y;
	state.p = stateOld->p;			// Processor flags
	state.ddr = stateOld->ddr;
	state.pr = stateOld->pr;		// Port
	state.pc = stateOld->pc;
	state.sp = stateOld->sp;
	state.intr[0] = stateOld->intr[0];		// Interrupt state
	state.intr[1] = stateOld->intr[1];
	state.intr[2] = stateOld->intr[2];
	state.intr[3] = stateOld->intr[3];
	state.nmi_state = stateOld->nmi_state;	
	state.dfff_byte = stateOld->dfff_byte;
	state.basic_in = stateOld->basic_in;
	state.kernal_in = stateOld->kernal_in;
	state.char_in = stateOld->char_in;
	state.io_in = stateOld->io_in;
	
	memcpy(RAM, p1, 0x8000);
	memcpy(IO_Ram, p1+0x8000, 0x1000);
	memcpy(RAM + 0x8000, p2, 0x8000);
	memcpy(Color, p2+0x8000, 0x400);
	TheCPU->SetState(&state);

	return ((UInt32) stateOld + sizeof(MOS6510StateOld) - (UInt32) p2);
}


/*
 *  Save 1541 state to snapshot
 *
 *  number of bytes written to p
 */

uint32 C64::Save1541State(uint8 *p)
{
	MOS6502State *state =  (MOS6502State *) ((UInt32) (p+0x803) & 0xfffffffc);
	TheCPU1541->GetState(state);

	memcpy(p, RAM1541, 0x800);

	return ((UInt32) state + sizeof(MOS6502State) - (UInt32) p);
}


/*
 *  Load 1541 state from snapshot
 */

uint16 C64::Load1541State(uint8 *p)
{
	MOS6502State *state = (MOS6502State *) ((UInt32) (p+0x803) & 0xfffffffc);

	memcpy(RAM1541, p, 0x800);
	TheCPU1541->SetState(state);

	return ((UInt32) state + sizeof(MOS6502State) - (UInt32) p);
}

uint16 C64::Load1541StateOld(uint8 *p)
{
	MOS6502State state;
	MOS6502StateOld *stateOld = (MOS6502StateOld *) ((UInt32) (p+0x803) & 0xfffffffc);

  memcpy(&state, stateOld, sizeof(MOS6502StateOld));
  state.ar = 0;
  state.ar2 = 0;
	state.state = 0;
	state.op = 0;
	state.rdbuf = 0;

	memcpy(RAM1541, p, 0x800);
	TheCPU1541->SetState(&state);

	return ((UInt32) stateOld + sizeof(MOS6502StateOld) - (UInt32) p);
}


/*
 *  Save VIC state to snapshot
 */

uint16 C64::SaveVICState(uint8 *p)
{
	MOS6569State *state = (MOS6569State *) (((UInt32) p+3) & 0xfffffffc);
	TheVIC->GetState(state);
	return ((UInt32) state + sizeof(MOS6569State) - (UInt32) p);
}


/*
 *  Load VIC state from snapshot
 */

uint16 C64::LoadVICState(uint8 *p)
{
	MOS6569State *state = (MOS6569State *) (((UInt32) p+3) & 0xfffffffc);

	TheVIC->SetState(state);
	return ((UInt32) state + sizeof(MOS6569State) - (UInt32) p);
}

uint16 C64::LoadVICStateOld(uint8 *p)
{
  int i;
  MOS6569State state;
	MOS6569StateOld *stateOld = (MOS6569StateOld *) (((UInt32) p+3) & 0xfffffffc);

	state.m0x = stateOld->m0x;
	state.m0y = stateOld->m0y;
	state.m1x = stateOld->m1x;
	state.m1y = stateOld->m1y;
	state.m2x = stateOld->m2x;
	state.m2y = stateOld->m2y;
	state.m3x = stateOld->m3x;
	state.m3y = stateOld->m3y;
	state.m4x = stateOld->m4x;
	state.m4y = stateOld->m4y;
	state.m5x = stateOld->m5x;
	state.m5y = stateOld->m5y;
	state.m6x = stateOld->m6x;
	state.m6y = stateOld->m6y;
	state.m7x = stateOld->m7x;
	state.m7y = stateOld->m7y;
	state.mx8 = stateOld->mx8;
	if(state.mx8 & 0x01) state.m0x += 256;
	if(state.mx8 & 0x02) state.m1x += 256;
	if(state.mx8 & 0x04) state.m2x += 256;
	if(state.mx8 & 0x08) state.m3x += 256;
	if(state.mx8 & 0x10) state.m4x += 256;
	if(state.mx8 & 0x20) state.m5x += 256;
	if(state.mx8 & 0x40) state.m6x += 256;
	if(state.mx8 & 0x80) state.m7x += 256;

	state.ctrl1 = stateOld->ctrl1;
	state.raster_y = stateOld->raster;
	state.lpx = stateOld->lpx;
	state.lpy = stateOld->lpy;
	state.me = stateOld->me;
	state.ctrl2 = stateOld->ctrl2;
	state.mye = stateOld->mye;
	state.vbase = stateOld->vbase;
	state.irq_flag = stateOld->irq_flag;
	state.irq_mask = stateOld->irq_mask;
	state.mdp = stateOld->mdp;
	state.mmc = stateOld->mmc;
	state.mxe = stateOld->mxe;
	state.clx_spr = stateOld->mm;
	state.clx_bgr = stateOld->md;

	state.ec = stateOld->ec;
	state.b0c = stateOld->b0c;
	state.b1c = stateOld->b1c;
	state.b2c = stateOld->b2c;
	state.b3c = stateOld->b3c;
	state.mm0 = stateOld->mm0;
	state.mm1 = stateOld->mm1;
	state.m0c = stateOld->m0c;
	state.m1c = stateOld->m1c;
	state.m2c = stateOld->m2c;
	state.m3c = stateOld->m3c;
	state.m4c = stateOld->m4c;
	state.m5c = stateOld->m5c;
	state.m6c = stateOld->m6c;
	state.m7c = stateOld->m7c;

	state.irq_raster = stateOld->irq_raster;		// IRQ raster line
	state.vc = stateOld->vc;				// Video counter
	state.vc_base = stateOld->vc_base;			// Video counter base
	state.rc = stateOld->rc;				// Row counter
	state.spr_dma = stateOld->spr_dma;			// 8 Flags: Sprite DMA active
	state.spr_disp = stateOld->spr_disp;			// 8 Flags: Sprite display active
	for(i=0; i<8; ++i)
	  state.mc[i] = stateOld->mc[i];			// Sprite data counters
	for(i=0; i<8; ++i)
  	state.mc_base[i] = stateOld->mc_base[i];		// Sprite data counter bases
	state.display_state = stateOld->display_state;		// true: Display state, false: Idle state
	state.bad_line = stateOld->bad_line;			// Flag: Bad Line state
	state.bad_line_enable = stateOld->bad_line_enable;	// Flag: Bad Lines enabled for this frame
	state.lp_triggered = stateOld->lp_triggered;		// Flag: Lightpen was triggered in this frame
	state.border_on = stateOld->border_on;			// Flag: Upper/lower border on (Frodo SC: Main border flipflop)
  state.frame_skipped = stateOld->frame_skipped;
  state.draw_this_line = false;
  
	state.bank_base = stateOld->bank_base;		// VIC bank base address
	state.matrix_base = stateOld->matrix_base;		// Video matrix base
	state.char_base = stateOld->char_base;		// Character generator base
	state.bitmap_base = stateOld->bitmap_base;		// Bitmap base

  state.spr_exp_y = ~state.mye;
	state.first_ba_cycle = 0;
	for(i=0; i<8; ++i)
  	state.sprite_base[i] = 0;	// Sprite bases
	state.cycle = 0x102;				// Current cycle in line (1..63)
	state.raster_x = (UInt16) (0xfffc + 8 * 53);		// Current raster x position
	state.ml_index = 0;			// Index in matrix/color_line[]
	state.ud_border_on = true;		// Flag: Upper/lower border on

	TheVIC->SetState(&state);
	return ((UInt32) stateOld + sizeof(MOS6569StateOld) - (UInt32) p);
}

/*
 *  Save SID state to snapshot
 */

uint16 C64::SaveSIDState(uint8 *p)
{
	MOS6581State *state = (MOS6581State *) (((UInt32) p+3) & 0xfffffffc);
	TheSID->GetState(state);
	return ((UInt32) state + sizeof(MOS6581State) - (UInt32) p);
}


/*
 *  Load SID state from snapshot
 */

uint16 C64::LoadSIDState(uint8 *p)
{
	MOS6581State *state = (MOS6581State *) (((UInt32) p+3) & 0xfffffffc);

	TheSID->SetState(state);
	return ((UInt32) state + sizeof(MOS6581State) - (UInt32) p);
}


/*
 *  Save CIA states to snapshot
 */

uint16 C64::SaveCIAState(uint8 *p)
{
	MOS6526State *state = (MOS6526State *) (((UInt32) p+3) & 0xfffffffc);
	TheCIA1->GetState(state);
	state = (MOS6526State *) (((UInt32) state + sizeof(MOS6526State) + 3) & 0xfffffffc);
	TheCIA2->GetState(state);

	return ((UInt32) state + sizeof(MOS6526State) - (UInt32) p);
}


/*
 *  Load CIA states from snapshot
 */

uint16 C64::LoadCIAState(uint8 *p)
{
	MOS6526State *state = (MOS6526State *) (((UInt32) p+3) & 0xfffffffc);

	TheCIA1->SetState(state);
	state = (MOS6526State *) (((UInt32) state + sizeof(MOS6526State) + 3) & 0xfffffffc);
	TheCIA2->SetState(state);

	return ((UInt32) state + sizeof(MOS6526State) - (UInt32) p);
}

uint16 C64::LoadCIAStateOld(uint8 *p)
{
	MOS6526State state;
	MOS6526StateOld *stateOld = (MOS6526StateOld *) (((UInt32) p+3) & 0xfffffffc);

  memcpy(&state, stateOld, sizeof(MOS6526StateOld));
  
	state.CyclesTillAction = 1;
	state.CyclesTillActionCnt = 1;
  state.has_new_cra = false;
  state.has_new_crb = false;
  state.new_cra = 0;
  state.new_crb = 0;
  state.ta_irq_next_cycle = false;
  state.tb_irq_next_cycle = false;
	state.ta_cnt_phi2 = ((state.cra & 0x20) == 0x00);
	state.tb_cnt_phi2 = ((state.crb & 0x60) == 0x00);
	state.tb_cnt_ta = ((state.crb & 0x60) == 0x40);
	state.ta_state = state.ta_cnt_phi2 ? T_COUNT : T_STOP;
	state.tb_state = (state.tb_cnt_phi2 || state.tb_cnt_ta) ? T_COUNT : T_STOP;
 
	TheCIA1->SetState(&state);

  // Same for CIA2
	stateOld = (MOS6526StateOld *) (((UInt32) stateOld + sizeof(MOS6526StateOld) + 3) & 0xfffffffc);
  memcpy(&state, stateOld, sizeof(MOS6526StateOld));
  
	state.CyclesTillAction = 1;
	state.CyclesTillActionCnt = 1;
  state.has_new_cra = false;
  state.has_new_crb = false;
  state.new_cra = 0;
  state.new_crb = 0;
  state.ta_irq_next_cycle = false;
  state.tb_irq_next_cycle = false;
	state.ta_cnt_phi2 = ((state.cra & 0x20) == 0x00);
	state.tb_cnt_phi2 = ((state.crb & 0x60) == 0x00);
	state.tb_cnt_ta = ((state.crb & 0x60) == 0x40);
	state.ta_state = state.ta_cnt_phi2 ? T_COUNT : T_STOP;
	state.tb_state = (state.tb_cnt_phi2 || state.tb_cnt_ta) ? T_COUNT : T_STOP;

	TheCIA2->SetState(&state);

	return ((UInt32) stateOld + sizeof(MOS6526StateOld) - (UInt32) p);
}


/*
 *  Save 1541 GCR state to snapshot
 */

uint16 C64::Save1541JobState(uint8 *p)
{
	Job1541State *state = (Job1541State *) (((UInt32) p+3) & 0xfffffffc);
	TheJob1541->GetState(state);
	return ((UInt32) state + sizeof(Job1541State) - (UInt32) p);
}


/*
 *  Load 1541 GCR state from snapshot
 */

uint16 C64::Load1541JobState(uint8 *p)
{
	Job1541State *state = (Job1541State *) (((UInt32) p+3) & 0xfffffffc);

	TheJob1541->SetState(state);
	return ((UInt32) state + sizeof(Job1541State) - (UInt32) p);
}

#define SNAPSHOT_1541 1


/*
 *  Save snapshot (emulation must be paused and in VBlank)
 *
 */
 
void C64::SaveSnapshot(uint8 *block1, uint8 *block2)
{
	uint8 *p1, *p2;
	uint8 flags;

	p1=block1;
	// SNAPSHOT_HEADER
	*p1++='F'; *p1++='r'; *p1++='o'; *p1++='d'; *p1++='o'; *p1++='S'; 
	*p1++='n'; *p1++='a'; *p1++='p'; *p1++='s'; *p1++='h'; *p1++='o'; 
	*p1++='t'; *p1++='\n';
	*p1++=SNAPSHOT_VERSION;	// Version number
	flags = 0;
	if (ThePrefs.Emul1541Proc)
		flags |= SNAPSHOT_1541;
	*p1++=flags;
	p1+=SaveVICState(p1);
	p1+=SaveSIDState(p1);
	p1+=SaveCIAState(p1);

	p2=block2;

	p2+=SaveCPUState(p1, p2);
	p1+=0x9000;

	if (ThePrefs.Emul1541Proc) 
	{
		//memcpy(p2, ThePrefs.DrivePath, 256);
		memset(p2, 0, 256); // After load of snapshot, we don't have a disk attached
		p2+=256;

		p2+=Save1541State(p2);
		p2+=Save1541JobState(p2);
	}
}


/*
 *  Load snapshot (emulation must be paused and in VBlank)
 */

bool C64::LoadSnapshot(uint8 *block1, uint8 *block2)
{
	uint8 *p1, *p2;
	unsigned char Header[14];
	uint8 flags;
	uint8 *vicptr;	// Pointer to VIC data
  bool bOldSnapshot = false;
  
	p1=Header;
	*p1++='F'; *p1++='r'; *p1++='o'; *p1++='d'; *p1++='o'; *p1++='S'; 
	*p1++='n'; *p1++='a'; *p1++='p'; *p1++='s'; *p1++='h'; *p1++='o';
	*p1++='t'; *p1++='\n';

	if (memcmp(Header, block1, sizeof(Header))) 
	{
		memcpy(Header, block1, sizeof(Header)-1);
		Header[sizeof(Header)-1]='\0';
		ShowRequester(REQ_BAD_SNAPSHOT, (uint32)Header,0,0,0);
		return false;
	}
	p1=block1+sizeof(Header);

	if (*p1 == SNAPSHOT_VERSION_OLD) 
	{
	  // Load snapshot from prev. version
		bOldSnapshot = true;
	}
	else if (*p1 != SNAPSHOT_VERSION) 
	{
		ShowRequester(REQ_BAD_SNAPSHOT_VERSION, *p1,0,0,0);
		return false;
	}
	p1++;
	flags = *p1++;
	vicptr = p1;
	p2=block2;

  if(bOldSnapshot)
  {
  	p1+=LoadVICStateOld(p1);
  	p1+=LoadSIDState(p1);
  	p1+=LoadCIAStateOld(p1);
  	p2+=LoadCPUStateOld(p1, p2);
  }
  else
  {
  	p1+=LoadVICState(p1);
  	p1+=LoadSIDState(p1);
  	p1+=LoadCIAState(p1);
  	p2+=LoadCPUState(p1, p2);
  }
	p1+=0x9000;

	if ((flags & SNAPSHOT_1541) != 0) 
	{
		// First switch on emulation
		TheNewPrefs = ThePrefs;
		//memcpy(TheNewPrefs.DrivePath, p2, 256);
		memset(TheNewPrefs.DrivePath, 0, 256); // No information about disk
		p2+=256;
		TheNewPrefs.Emul1541Proc = true;
		NewPrefs(&TheNewPrefs);
		ThePrefs = TheNewPrefs;
	
		// Then read the context
    if(bOldSnapshot)
    {
  		p2+=Load1541StateOld(p2);
  		p2+=Load1541JobState(p2);
    }
    else
    {
  		p2+=Load1541State(p2);
  		p2+=Load1541JobState(p2);
    }
	} 
	else if (ThePrefs.Emul1541Proc) 
	{	// No emulation in snapshot, but currently active?
		TheNewPrefs = ThePrefs;
		TheNewPrefs.Emul1541Proc = false;
		NewPrefs(&TheNewPrefs);
		ThePrefs = TheNewPrefs;
	}

	p1=vicptr;
  if(bOldSnapshot)
  	p1+=LoadVICStateOld(p1);	// Load VIC data twice in SL (is REALLY necessary sometimes!)
  else
  	p1+=LoadVICState(p1);	// Load VIC data twice in SL (is REALLY necessary sometimes!)
  
	return true;
}


/*
 *  Constructor, system-dependent things
 */

void C64::c64_ctor1(void)
{
	RAM = armletData->ram;
  Basic = armletData->basicROM;
  Kernal = armletData->kernalROM;
  Char = armletData->charROM;
  Color = armletData->color;
  RAM1541 = armletData->ram1541;
  ROM1541 = armletData->rom1541;
  IO_Ram = armletData->io_ram;

	if (armletData->cmd & CMD_GETPREFS)
		ThePrefs.GetState(armletData->prefs);
	if (armletData->cmd & CMD_SETPREFS) 
	{
		ThePrefs.SetState(armletData->prefs);
		armletData->cmd &= ~(CMD_SETPREFS);
	}
}


/*
 *  Start emulation
 */
void C64::Run(void)
{
	// Reset chips
	TheCPU->Reset();
	TheSID->Reset();
	TheCIA1->Reset();
	TheCIA2->Reset();
	TheCPU1541->Reset();

	// Patch kernal IEC routines
	orig_kernal_1d84 = Kernal[0x1d84];
	orig_kernal_1d85 = Kernal[0x1d85];
	PatchKernal(ThePrefs.FastReset, ThePrefs.Emul1541Proc);

	// Start the CPU thread
	thread_running = true;
	quit_thyself = false;
	have_a_break = false;

	if (!armletData->errCode) 
		thread_func();
}


/*
 *  Stop emulation
 */
void C64::Quit(void)
{
	// Ask the thread to quit itself if it is running
	if (thread_running) 
	{
		ShowRequester(REQ_DEBUG, 0xbaddead, 1,0,0);
		quit_thyself = true;
		thread_running = false;
	}
}


/*
 *  Pause emulation
 */
void C64::Pause(void)
{
	TheSID->PauseSound();
}


/*
 *  Resume emulation
 */
void C64::Resume(void)
{
	TheSID->ResumeSound();
}


/*
 *  Vertical blank: Poll keyboard and joysticks, update window
 */
void C64::VBlank(bool draw_frame)
{
	/* Jump back to 68k code.
	   This handles events, timing, redrawing, sound playback...
	   All results are stuffed into appropriate fields in armletData.
	   This is the only place the armlet calls the 68k program, except
	   ShowRequester, new and delete.
	*/
	//armletData->drawFrame=draw_frame;
	PutR10(armletData->r10org);
  armletData->r9curr=GetR9();
	PutR9(armletData->r9org);
	(armletData->call68k) (armletData->emulStateP, (unsigned long)armletData->vBlank, &armletData, 4);
	armletData->r10org=GetR10();
	PutR10((uint32)armletData);
	armletData->r9org=GetR9();
  PutR9(armletData->r9curr);

	// Commands
	if (armletData->cmd & CMD_GETPREFS)
		ThePrefs.GetState(armletData->prefs);
	if (armletData->cmd & CMD_SETPREFS) 
	{
		TheNewPrefs.SetState(armletData->prefs);
		NewPrefs(&TheNewPrefs);
		ThePrefs = TheNewPrefs;
		armletData->cmd &= ~(CMD_SETPREFS);
	}
	if (armletData->cmd & CMD_LOAD) 
	{
		LoadSnapshot(armletData->snapshot1, armletData->snapshot2);
		armletData->cmd &= ~(CMD_LOAD);
	}
	if (armletData->cmd & CMD_SAVE)
		SaveSnapshot(armletData->snapshot1, armletData->snapshot2);
	if (armletData->cmd & CMD_RESET) 
	{
		Reset();
		armletData->cmd &= ~(CMD_RESET);
	}

	// Poll keyboard
	TheDisplay->PollKeyboard(TheCIA1->KeyMatrix, TheCIA1->RevMatrix);

	// Poll joysticks
	TheCIA1->Joystick1 = poll_joystick(0);
	TheCIA1->Joystick2 = poll_joystick(1);

	if (ThePrefs.JoystickSwap) 
	{
		uint8 tmp = TheCIA1->Joystick1;
		TheCIA1->Joystick1 = TheCIA1->Joystick2;
		TheCIA1->Joystick2 = tmp;
	}

	TheCIA1->UpdateDataPorts();
	
	// Count TOD clocks
	TheCIA1->CountTOD();
	TheCIA2->CountTOD();
}


/*
 *  Poll joystick port, return CIA mask
 */
uint8 C64::poll_joystick(int port)
{
	uint8 joy_state=0xff;
	
	if (port==0 && !armletData->prefs->Joystick1On)
		return 0xff;
	if (port==1 && !armletData->prefs->Joystick2On)
		return 0xff;
		
	if (armletData->joystick & JOYFIRE)
		joy_state &= 0xef;
	if (armletData->joystick & JOYRIGHT)
		joy_state &= 0xf7;
	if (armletData->joystick & JOYLEFT)
		joy_state &= 0xfb;
	if (armletData->joystick & JOYDOWN)
		joy_state &= 0xfd;
	if (armletData->joystick & JOYUP)
		joy_state &= 0xfe;

	return joy_state;
}


/*
 * The emulation's main loop
 */
void C64::thread_func(void)
{
	while (!quit_thyself) 
	{
		if(ThePrefs.SingleCycleEmulation)
		{
  		TheVIC->EmulateLineSC();
    }
    else
    {
  		// The order of calls is important here
  		int cycles = TheVIC->EmulateLine();
  		if(ThePrefs.SIDOn)
  		  TheSID->EmulateLine();
#if !PRECISE_CIA_CYCLES
  		TheCIA1->EmulateLine(ThePrefs.CIACycles);
  		TheCIA2->EmulateLine(ThePrefs.CIACycles);
#endif
  		if (ThePrefs.Emul1541Proc) 
  		{
  			int cycles_1541 = ThePrefs.FloppyCycles;
  			TheCPU1541->CountVIATimers(cycles_1541);
  
  			if (!TheCPU1541->Idle) 
  			{
  				TheCPU1541->EmulateLine(cycles_1541, cycles); // EmulateLine of CPUC64 called in there
  			} 
  			else
  				TheCPU->EmulateLine(cycles);
  		} 
  		else 
  		{
  			// 1541 processor disabled, only emulate 6510
  			TheCPU->EmulateLine(cycles);
  		}
    }

    if(SwitchToSC)
    {
      // Single cycle emulation switched on
    	SwitchToSC = false;
    	TheCIA1->SwitchToSC();
    	TheCIA2->SwitchToSC();
      TheVIC->SwitchToSC();
      TheCPU->SwitchToSC();
      if(ThePrefs.Emul1541Proc)
        TheCPU1541->SwitchToSC();
    }
    
    if(SwitchToStandard)
    {
      // Single cycle emulation switched off
      SwitchToStandard = false;
    	TheCIA1->SwitchToStandard();
    	TheCIA2->SwitchToStandard();
      TheVIC->SwitchToStandard();
      TheCPU->SwitchToStandard();
      if(ThePrefs.Emul1541Proc)
        TheCPU1541->SwitchToStandard();
    }
    
		if (armletData->errCode)
			quit_thyself=1;
	}
	if (armletData->errCode==0)
		ShowRequester(REQ_DEBUG, 0xbaddead, 2,0,0);
}


uint8 C64::Random(void)
{
	seed = seed * 1103515245L + 12345;
  return (seed >> 16)&0xff;
}

void C64::SeedRandom(uint32 s)
{
	seed=s;
}
