/*
 * PalmOSARMLET.c
 * Procedures to initialize and release armlet code and data
 */
#define _PALMOSARMLET_C

#include <PalmOS.h>
#include <Bitmap.h>
#include <MemoryMgr.h>
#include <DataMgr.h>

#include "PalmOSGUI.h"
#include "PalmOSFileIO.h"
#include "PalmOSArmlet.h"
#include "FrodoResources.h"
#include "PalmHelper.h"
#include "PalmHardware.h"


UInt8 *armlet=NULL;						// Pointer to armlet code
ArmletData *armletData=NULL;	// Pointer to armlet data
BitmapType *RAMbmp = NULL;		// Pointer to bitmap which is used to get a 64K block of memory for RAM
Boolean IsNuked = false;


//***********************************************************
// Procedure: CreateArmlet
//
// Parameter: -
//
// Returnval: Errorcode or 0 if successful
//
// Description: Load armletcode into memory.
//***********************************************************
#define MAXCHUNKS 5
UInt32 CreateArmlet()
{
	MemHandle armletPart[MAXCHUNKS] = { 0, 0, 0, 0, 0 };
	MemPtr p;
	UInt32 armletSize, offs;
	UInt32 nParts;
	UInt32 i;
	UInt32 error;

	// Reassemble armlet from resources and stuff it in feature memory
	// Armlet is stored as Arml IDX_ARMLET, +1, +2, +3, ...
	armletSize = 0;
	for (i=0; i<MAXCHUNKS; i++) 
	{
		armletPart[i] = GetResource(armletType, IDX_ARMLET+i);
		if (!armletPart[i]) 
			break;
		armletSize += MemHandleSize(armletPart[i]);
	}
	if (i>=MAXCHUNKS) 
	{
	  ShowErrorResLong(StringErrArmletToBig, i);
		for (i=0; i<MAXCHUNKS; i++) 
		  FreeResource(armletPart[i]);
		return 1;
	}
	if (i==0) 
		return DmGetLastErr();
	nParts = i;
	
	// Free feature 0 just in case we exited without freeing it last time
	FtrPtrFree(appSignature, 0);
	error = FtrPtrNew(appSignature, 0, armletSize, (void **) &armlet);
	if (error) 
	{
		ShowErrorRes(StringErrNoFeatureForArmlet);
		for (i=0; i<nParts; i++) 
			FreeResource(armletPart[i]);
		return error;
	}

	offs = 0;
	// Write the parts to feature memory
	for (i=0; i<nParts; i++) 
	{
		p = MemHandleLock(armletPart[i]);
		error = DmWrite(armlet, offs, p, MemHandleSize(armletPart[i]));
		MemHandleUnlock(armletPart[i]);
		if (error) 
		{
		  ShowErrorResLong(StringErrFeatureWriteFailed, i);
			FtrPtrFree(appSignature, 0);
			for (i=0; i<nParts; i++)
				FreeResource(armletPart[i]);
			return error;
		}
		offs += MemHandleSize(armletPart[i]);
	}
	for (i=0; i<nParts; i++)
	  FreeResource(armletPart[i]);
	return 0;
}


//***********************************************************
// Procedure: CreateArmletData
//
// Parameter: -
//
// Returnval: Errorcode or 0 if successful
//
// Description: Allocate ArmletData and fill it in (mind the endianness!)
//***********************************************************
UInt32 CreateArmletData()
{
	UInt32 x;
	UInt32 i;
	UInt16 iDummy, iPrefSize;
	Int16 iResult;
	
	// Get and initialize memory for armlet data
	armletData = (ArmletData *) MemPtrNew(sizeof(ArmletData));
	if (!armletData) 
		return memErrNotEnoughSpace;
	MemSet(armletData, sizeof(ArmletData), 0x0);
	
	// Callbacks, needed some workarounds to get the pointers in there.
	x = (UInt32) VBlank68k;
	armletData->vBlank    = (void *) swap32(x);
	x = (UInt32) ShowRequester68k;
	armletData->requester = (void *) swap32(x);
	x = (UInt32) Malloc68k;
	armletData->malloc    = (void *) swap32(x);
	x = (UInt32) Free68k;
	armletData->free      = (void *) swap32(x);
	x = (UInt32) DmWrite68k;
	armletData->dmwrite   = (void *) swap32(x);
  
	// RAM & ROM
  RAMbmp = BmpCreate(256, 256, 8, NULL, &iDummy); // We use a 256x256 bitmap to get a 64K block of memory
	if (!RAMbmp) 
		return memErrNotEnoughSpace;
	armletData->ram = (MemPtr) swap32((UInt32) BmpGetBits(RAMbmp));
	if (!armletData->ram)
	{
		BmpDelete(RAMbmp);
		RAMbmp = NULL;
		return memErrNotEnoughSpace;
  }

	armletData->kernalROM = CopyRes(romType, IDX_KERNAL, "KernalROM");
	if (!armletData->kernalROM) 
		return memErrNotEnoughSpace;
		
	armletData->basicROM = GetRes(romType, IDX_BASIC, "BasicROM");
	if (!armletData->basicROM) 
		return memErrNotEnoughSpace;
		
	armletData->charROM = GetRes(romType, IDX_CHAR, "CharROM");
	if (!armletData->charROM) 
		return memErrNotEnoughSpace;
		
	armletData->rom1541 = CopyRes(romType, IDX_1541, "1541ROM");
	if (!armletData->rom1541) 
		return memErrNotEnoughSpace;
	
	armletData->color = (UInt8 *) NewSwapped(0x0400); // Color RAM
	if (!armletData->color) 
		return memErrNotEnoughSpace;
		
	armletData->ram1541 = (UInt8 *) NewSwapped(0x0800); // Diskdrive RAM
	if (!armletData->ram1541) 
		return memErrNotEnoughSpace;
		
	armletData->io_ram = (UInt8 *) NewSwapped(0x1000); // RAM to simulate IO (only if CPU jmp to $dxxx occurs)
	if (!armletData->io_ram) 
		return memErrNotEnoughSpace;

	// Various tables and precalculations
	armletData->expTable = GetRes(tblType, IDX_EXPTABLE, "ExpTable");
	if (!armletData->expTable) 
		return memErrNotEnoughSpace;
		
	armletData->multiExpTable=GetRes(tblType, IDX_MULTIEXPTABLE, "MultiExpTable");
	if (!armletData->multiExpTable) 
		return memErrNotEnoughSpace;
		
	armletData->textColorTable=GetRes(tblType, IDX_TEXTCOLORTABLE, "Text color lookup table");
	if (!armletData->textColorTable) 
		return memErrNotEnoughSpace;
		
	armletData->triTable=GetRes(tblType, IDX_TRITABLE, "TriTable");
	if (!armletData->triTable) 
		return memErrNotEnoughSpace;
		
	armletData->triSawTable=GetRes(tblType, IDX_TRISAWTABLE, "TriSawTable");
	if (!armletData->triSawTable) 
		return memErrNotEnoughSpace;
		
	armletData->triRectTable=GetRes(tblType, IDX_TRIRECTTABLE, "TriRectTable");
	if (!armletData->triRectTable) 
		return memErrNotEnoughSpace;
		
	armletData->sawRectTable=GetRes(tblType, IDX_SAWRECTTABLE, "SawRectTable");
	if (!armletData->sawRectTable) 
		return memErrNotEnoughSpace;
		
	armletData->triSawRectTable=GetRes(tblType, IDX_TRISAWRECTTABLE, "TriSawRectTable");
	if (!armletData->triSawRectTable) 
		return memErrNotEnoughSpace;
		
	armletData->egTable=GetRes(tblType, IDX_EGTABLE, "EGTable");
	if (!armletData->egTable) 
		return memErrNotEnoughSpace;
		
	armletData->egdrShift=GetRes(tblType, IDX_EGDRSHIFT, "EGDRShift");
	if (!armletData->egdrShift) 
		return memErrNotEnoughSpace;
		
	armletData->sampleTab=GetRes(tblType, IDX_SAMPLETAB, "SampleTab");
	if (!armletData->sampleTab) 
		return memErrNotEnoughSpace;
		
	armletData->sintab=GetRes(tblType, IDX_SINTAB, "SinTab");
	if (!armletData->sintab) 
		return memErrNotEnoughSpace;
		
	armletData->resonanceLP=GetRes(tblType, IDX_RESONANCELP, "ResonanceLP");
	if (!armletData->resonanceLP) 
		return memErrNotEnoughSpace;
		
	armletData->resonanceHP=GetRes(tblType, IDX_RESONANCEHP, "ResonanceHP");
	if (!armletData->resonanceHP) 
		return memErrNotEnoughSpace;
  
	armletData->modetab=GetRes(tblType, IDX_MODETAB, "ModeTab");
	if (!armletData->modetab) 
		return memErrNotEnoughSpace;

	armletData->optab=GetRes(tblType, IDX_OPTAB, "OpTab");
	if (!armletData->optab) 
		return memErrNotEnoughSpace;

	// 1541 error messages
	for (i=0; i<12; i++) 
	{
		armletData->errors1541[i] = GetRes(strType, IDX_ERRORS1541+i,	"Errors1541");
		if (!armletData->errors1541[i])
			return memErrNotEnoughSpace;
	}
	
	// Snapshot buffers
	if (!(armletData->snapshot1 = NewSwapped(SNAPSHOT_SIZE_1)))
		return memErrNotEnoughSpace;
	if (!(armletData->snapshot2 = NewSwapped(SNAPSHOT_SIZE_2)))
		return memErrNotEnoughSpace;
	
  // Load Palm specific preferences
  iDummy = sizeof(struct PreferencesPalm);
  prefsPalmData.SpeedLimit = 100; // New parameter -> set default for old version of PalmPrefs
  iResult = PrefGetAppPreferences(appSignature, IDX_PALMPREFS, &prefsPalmData, &iDummy, false);
  if(iResult == noPreferenceFound)
  {
    // No prefs -> first call
    MemSet(&prefsPalmData, sizeof(struct PreferencesPalm), 0);

    // Create default commands
    StrCopy(prefsPalmData.C64CommandTitle[0], "Directory");
    StrCopy(prefsPalmData.C64Command[0], "LOAD\"$\",8\nLIST\n");

    StrCopy(prefsPalmData.C64CommandTitle[1], "Load/Run");
    StrCopy(prefsPalmData.C64Command[1], "LOAD\":*\",8,1\nRUN\n");

    StrCopy(prefsPalmData.C64CommandTitle[2], "Softreset");
    StrCopy(prefsPalmData.C64Command[2], "SYS64738\n");
  }
  prefsPalm = &prefsPalmData;
  
	// Set up prefs
	// Normally we want the Armlet to set it's prefs based on the saved ones
	armletData->cmd |= CMD68K_SETPREFS;
  iPrefSize = 0;
  iResult = PrefGetAppPreferences(appSignature, IDX_PREFS, NULL, &iPrefSize, false);
  iDummy = sizeof(struct PrefsState);
  iResult = PrefGetAppPreferences(appSignature, IDX_PREFS, &prefsData, &iDummy, false);
	if (iResult == noPreferenceFound) 
	{
		MemSet(&prefsData, sizeof(struct PrefsState), 0);
		IsNuked = true;
	}
	else if(iPrefSize != sizeof(struct PrefsState))
	{
	  // Found old prefs -> init new members
	  prefsData.SingleCycleEmulation = 0;
	  prefsData.SIDOn = 0;
	}
	prefs = &prefsData;
	
	// store a normal-endian pointer in prefs
	armletData->prefs = (struct PrefsState *) swap32((UInt32) prefs);
	if (IsNuked) 
	{
		// Tell armlet to give us the default prefs if first time or Frodo crashed
		armletData->cmd |=   CMD68K_GETPREFS;
		armletData->cmd &= ~(CMD68K_SETPREFS);
	} 
	else 
	{
		// Only set the key mask if we have prefs, or all keys will be blocked!
		SetKeyMask();
	}
	
	// After loading snapshot at startup, we have never a disk in drive
	prefs->DriveType = 0;
	MemSet(prefs->DrivePath, 32, 0);
	armletData->disk = 0;
	armletData->diskType = 0;
  armletData->track1541 = 0;
  armletData->driveLED = 0;
  
	if (IsNuked)
	{
		// 'Nuked' prefs is a good indicator we are running Frodo for the first time.
		firstTime = true;

    // First time -> set default values
  	prefsPalm->Volume = 0;
  	//prefsPalm->Volume = 16; // 1%, 100% it is very nasty when new version is loaded to device // 1024; // 100%
		prefs->SIDOn = false;
		prefs->AdaptiveFrameSkip = 0;
		prefs->SkipFrames = swap32(1);
  	
		// Start without single cycle emulation
		prefs->SingleCycleEmulation = 0;

  	// Check if we have 5-way keys
  	if(DeviceHas5Way())
  	{
  		// Assign 5-way to joystick
	  	prefsPalm->KeyUp = 0x2;
	  	prefsPalm->KeyLeft = 0x1000000;
	  	prefsPalm->KeyRight = 0x2000000;
	  	prefsPalm->KeyDown = 0x4;
	  	prefsPalm->KeyFire = 0x40;
  	}
  	else
  	{
	  	prefsPalm->KeyUp = 0;
	  	prefsPalm->KeyLeft = 0;
	  	prefsPalm->KeyRight = 0;
	  	prefsPalm->KeyDown = 0;
	  	prefsPalm->KeyFire = 0;
		}
  	prefsPalm->JoyPen = JOYPEN_GRAFFITI;
  	if(is320)
  	 	prefsPalm->ScreenMode = SCREENMODE_256;
  	else
  	 	prefsPalm->ScreenMode = SCREENMODE_256_BUFFERED;
  	prefsPalm->SaveToSDCard = SAVE2SD_NEVER;
  	for(i=0; i<66; ++i)
  	  prefsPalm->KeyMapping[i] = 0;
    MemMove(prefsPalm->colorTable, colorTablePreset, 16 * sizeof(RGBColorType));
		prefsPalm->SoundBufferSize = FRAME_SAMPLES * 2;
		prefsPalm->SpeedLimit = 100;
		
		// get snapshot defaults from variables
		prefsPalm->lastSnapshotIndex = lastIndex;
		MemSet(prefsPalm->lastSnapshotName, sizeof(prefsPalm->lastSnapshotName), 0);
		StrNCopy(prefsPalm->lastSnapshotName, lastName, 31);
		
    SavePrefsPalm();
  } 
  else 
  {
		//not nuked prefs - load snapshot variables from prefs
		lastIndex = prefsPalm->lastSnapshotIndex;
		MemSet(lastName, sizeof(lastName), 0);
		StrNCopy(lastName, prefsPalm->lastSnapshotName, 31);
	}
	
	return 0;
}


//***********************************************************
// Procedure: FreeArmletData
//
// Parameter: -
//
// Returnval: -
//
// Description: Release memory of armlet data.
//***********************************************************
void FreeArmletData(void)
{
	UInt32 i;
	
	if (!armletData) 
		return;

	// Free up everything...
	// Things that where allocated by MemPtrNew or CopyRes should get passed to MemPtrFree,
	// things that where created by GetResource should have their handles unlocked.
	// It is SUPREMELY important that allocs and frees match!

	// Release memory of C64
	if(NULL != RAMbmp)
	  BmpDelete(RAMbmp);	// Release RAM
	UncopyRes(armletData->kernalROM);
	UngetRes(armletData->basicROM);
	UngetRes(armletData->charROM);
	UncopyRes(armletData->rom1541);
	FreeSwapped(armletData->color);
	FreeSwapped(armletData->ram1541);
	FreeSwapped(armletData->io_ram);

	UngetRes(armletData->expTable);
	UngetRes(armletData->multiExpTable);
	UngetRes(armletData->textColorTable);
	UngetRes(armletData->triTable);
	UngetRes(armletData->triSawTable);
	UngetRes(armletData->triRectTable);
	UngetRes(armletData->sawRectTable);
	UngetRes(armletData->triSawRectTable);
	UngetRes(armletData->egTable);
	UngetRes(armletData->egdrShift);
	UngetRes(armletData->sampleTab);
	UngetRes(armletData->sintab);
	UngetRes(armletData->resonanceLP);
	UngetRes(armletData->resonanceHP);
	UngetRes(armletData->modetab);
	UngetRes(armletData->optab);

	for (i=0; i<12; i++)
		UngetRes(armletData->errors1541[i]);

	FreeSwapped(armletData->snapshot1);
	FreeSwapped(armletData->snapshot2);

	// Finally release memory of armlet data
	MemPtrFree(armletData);
	armletData = NULL;
}

