 /*
 * PalmOSGUI.c
 * The GUI for the Palm. This source is 68k, unlike all the other files
 * which are compiled for ARM.
 */
#define _PALMOSGUI_C

#include <PalmOS.h>
#include <Bitmap.h>
#include <MemoryMgr.h>
#include <DataMgr.h>

#include "PalmOSGUI.h"
#include "PalmOSArmlet.h"
#include "PalmOSFileIO.h"
#include "FrodoResources.h"
#include "PalmHelper.h"
#include "PalmHardware.h"
#include "PalmColorPresets.h"
#include "PalmC64Commands.h"


static Boolean AppHandleEvent(EventPtr e);
static Boolean HandleEventMain(EventPtr e);
static Boolean HandleEventColors(EventPtr e);
static Boolean HandleEventFile(EventPtr e);
static Boolean HandleEventGfxSound(EventPtr e);
static Boolean HandleEventJoystick(EventPtr e);
static Boolean HandleEventKeyboard(EventPtr e);
static Boolean HandleEventSampleKey(EventPtr e);
static UInt16 ShowDialog(UInt16);


struct PrefsState *prefs = NULL;	// Pointer to preferences of Frodo, data in preferences are in armlet endian order
struct PreferencesPalm *prefsPalm = NULL;	// Pointer to Palm specific settings
struct PrefsState prefsData;	// Pointer to preferences of Frodo, data in preferences are in armlet endian order
struct PreferencesPalm prefsPalmData;	// Pointer to Palm specific settings


// Capabilities
UInt8 canSound=0;	// Device has sound
UInt8 can256=0;		// Device can display 256 colors
UInt8 is320=0;		// At least 320 pixel with double density
UInt32 screenW=0;	// Width of screen
UInt32 screenH=0;	// Height of screen


// Various
Boolean crashed = false;		// Flag if Frodo crashed during last execution
Boolean firstTime = false;	// Flag if Frodo is running the first time
Boolean exiting = false;		// Flag if Frodo currently quits
Boolean pausedStateValid = false; // If pause mode was restored properly
Boolean pause = false;			// Flag if Frodo is paused
#define NSAVE2SD_MODE 3
char *save2SDList[NSAVE2SD_MODE] = { "Never", "Ask", "Always" };
Boolean bLaunchedFromSD = false;  // Flag if Frodo was launched from SD card


// Keys
Boolean keyStateChanged = false;	// Flag if user does something -> restart auto off timer
UInt32 g_lLastKeyState = 0;	// Keystate of last call of CheckKeyKey
UInt32 lKeyMask = 0;	// Original key mask
UInt32 usedKeys = 0;	// Mask with all currently used (mapped) Palm-keys
UInt16 inKey = 0;			// Current key of onscreen keyboard
UInt8 cmdLock = 0;		// Currently Commodore key locked on onscreen keyboard
UInt8 shiftLock = 0;	// Currently shift key locked on onscreen keyboard
UInt8 lowerCase = 0;	// Lower case active on onscreen keyboard
UInt32 ticksLastKeyDown = 0;	// Tickcount when Graffiti key down occured
UInt32 savedGraffitiKey = 0;	// Last sent Graffiti key
UInt32 dlgUp;			// Palm key used for up if configuration dialog is active
UInt32 dlgLeft;   // Palm key used for left if configuration dialog is active
UInt32 dlgRight;  // Palm key used for right if configuration dialog is active
UInt32 dlgDown;   // Palm key used for down if configuration dialog is active
UInt32 dlgFire;   // Palm key used for fire if configuration dialog is active
UInt32 sampledKey = 0;	// Pressed key in SampleKey dialog
UInt16 keyboardBmp = KeyboardBitmap;	// Current bitmap of keyboard
UInt32 oldJoy = 0xffffffffL;	// Last state of joystick
UInt16 joyBmp = 0;						// Current bitmap of joystick
Int16 g_lColPos[] = { 3, 10, 40, 47, 77, 84, 114, 126, 153 }; // Positions of columns for key mapping dialog

#define SENDKEYSBUFFERSIZE 128
Char strSendKeysBuffer[SENDKEYSBUFFERSIZE];
Char *pSendKeysBufferPos;
Boolean bSendKeysKeyDown = false;

/* C64 keyboard matrix:

    Bit 7   6   5   4   3   2   1   0
  0    CUD  F5  F3  F1  F7 CLR RET DEL
  1    SHL  E   S   Z   4   A   W   3
  2     X   T   F   C   6   D   R   5
  3     V   U   H   B   8   G   Y   7
  4     N   O   K   M   0   J   I   9
  5     ,   @   :   .   -   L   P   +
  6     /   ^   =  SHR HOM  ;   *   
  7    R/S  Q   C= SPC  2  CTL  <-  1

  Palm key codes are simply the position on the keyboard
0-16:   <- 1  2  3  4  5  6  7  8  9  0  +  -  P  HOM DEL F1
17-32: CTR Q  W  E  R  T  Y  U  I  O  P  @  *  ^  RST F3
33-48: R/S SL A  S  D  F  G  H  J  K  L  :  ;  =  RET F5
49-64: C= SHL Z  X  C  V  B  N  M  ,  .  / SHRCUD CLR F7 
65:             SPACE

  Tables for key translation
  Bit 0..2: row/column in C64 keyboard matrix
  Bit 3   : implicit shift
  Bit 5   : joystick emulation (bit 0..4: mask)
*/

const int key_byte[66] = {
	7, 7,  7, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6,  0,    0,
	   7,  7, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, -1,    0,
	   7, -1, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6,  0,    0,
	       7, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6,  0, 0, 0,
	                      7
};

const int key_bit[66] = {
	1, 0,  3, 0, 3, 0, 3, 0, 3, 0, 3, 0, 3, 0, 3,  0,    4,
	   2,  6, 1, 6, 1, 6, 1, 6, 1, 6, 1, 6, 1, 6, -1,    5,
	   7, -1, 2, 5, 2, 5, 2, 5, 2, 5, 2, 5, 2, 5,  1,    6,
	       5, 7, 4, 7, 4, 7, 4, 7, 4, 7, 4, 7, 4,  7, 2, 3,
	                      4,
};
/*
  Palm key codes are simply the position on the keyboard
0-16:   <- 1  2  3  4  5  6  7  8  9  0  +  -  P  HOM DEL F1
17-32: CTR Q  W  E  R  T  Y  U  I  O  P  @  *  ^  RST F3
33-48: R/S SL A  S  D  F  G  H  J  K  L  :  ;  =  RET F5
49-64: C= SHL Z  X  C  V  B  N  M  ,  .  / SHRCUD CLR F7 
65:             SPACE
*/
const uint8 C64KeyToDisplay[66] = {
  36, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 37, 38, 39, 54, 55, 49,
  56, 26, 32, 14, 27, 29, 34, 30, 18, 24, 25, 40, 41, 42, 57, 50,
  58, 59, 10, 28, 13, 15, 16, 17, 19, 20, 21, 43, 44, 45, 60, 51,
  61, 62, 35, 33, 12, 31, 11, 23, 22, 46, 47, 48, 63, 64, 65, 52, 53 };

// ASCII to C=64 key number conversion table for ASCII values 0-127
// +128 indicates shifted key
UInt8 C64KeyNumberFromAscii[129] ={
//  0:     NUL     SOH     STX     ETX     EOT     ENQ     ACK     BEL
	   255,    255,    255,    255,    255,    255,    255,    255,
//  8:     BS      TAB     LF      VT      FF      CR      SO      SI
	    15,    255,     47,    255,    255,     47,    255,    255,
// 16:     DLE     DC1     DC2     DC3     DC4     NAK     SYN     ETB
	   255,    255,    255,    255,    255,    255,    255,    255,
// 24:     CAN     EM      SUB     ESC     FS,<-   GS,->   RS,^    US,v
	   255,    255,    255,    255,  128+63,    63, 128+62,    62,
// 32:     ' '      !       "       #       $       %       &       '
	    65,  1+128,  2+128,  3+128,  4+128,  5+128,  6+128,  7+128,
// 40:      (       )       *       +       ,       -       .       /
	 8+128,  9+128,     29,     11,     58,     12,     59,     60,
// 48:      0       1       2       3       4       5       6       7
	    10,      1,      2,      3,      4,      5,      6,      7,
// 56:      8       9       :       ;       <       =       >       ?
	     8,      9,     44,     45, 58+128,     46, 59+128, 60+128,
// 64:      @       A       B       C       D       E       F       G
	    28,     35,     55,     53,     37,     20,     38,     39,
// 72:      H       I       J       K       L       M       N       O
	    40,     25,     41,     42,     43,     57,     56,     26,
// 80:      P       Q       R       S       T       U       V       W
	    27,     18,     21,     36,     22,     24,     54,     19,
// 88:      X       Y       Z       [       \       ]       ^       _
	    52,     23,     51, 44+128,    255, 45+128,    255,    255,
// 96:             a       b       c       d       e       f       g
	   255,     35,    55,     53,     37,     20,     38,     39,
//104:      h       i       j       k       l       m       n       o
	    40,     25,     41,     42,     43,     57,     56,     26,
//112:      p       q       r       s       t       u       v       w
	    27,     18,     21,     36,     22,     24,     54,     19,
//120:      x       y       z       {       |       }       ~      DEL
	    52,     23,     51,    255,    255,    255,    255,    15,
//128:      
     4+128 // mapping of  to $
};

// Clear text for Palm hardware keys
char *keys[33]={
	"Power", "PgUp",	"PgDn",	"Key1",	"Key2",	"Key3",	"Key4",	"Cradle",
	"Antenna", "Contrast", "Unknown1", "Unknown2", "Unknown3", "Unknown4", "Unknown5", "Unknown6",
	"Unknown7", "Unknown8", "Unknown9", "Unknown10", "Unknown11", "Unknown12", "Unknown13", "Unknown14",
	"Left", "Right", "Select", "Unknown18", "Unknown19", "Unknown20", "Unknown21", "Unknown22",
	"Unset"
};


// Coordinates and state of pen
Int32 penX;
Int32 penY;
Boolean penDown = false;


// Forms
Boolean doFileForm = false;		// Flag to open Filemanager form in next VBlank
Boolean doSaveAs = false;			// Flag to open SaveAs dialog in next VBlank
Boolean doQuickload = false;	// Flag to do Quickload in next VBlank
Boolean doQuicksave = false;	// Flag to do Quicksave in next VBlank
Boolean doFileLoadAfterSnap = false;  // Flag to load selected disk after loading a snapshot


// Fonts
MemHandle hFontStandard = NULL;
FontType *pFontStandard = NULL;
FontID lOldFont = 0;


// Graphics
UInt8 *screenBits = NULL; 	// Pointer to graphics data in direct mode
UInt8 *bmpBits = NULL;			// Pointer to graphics data in buffered mode
BitmapType *bmp = NULL;			// Bitmap in buffered mode
BitmapTypeV3 *bmpv3 = NULL;	// Bitmap in buffered mode
MemHandle joyHdl[10] = {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
BitmapPtr joyPtr[10] = {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
#define NSKIPLIST 6
char *frameskipList[NSKIPLIST] = { "Adaptive", "50",	"25",	"17",	"13",	"10" };
#define NSCREENMODE 2
char *screenmodeList[NSCREENMODE] = {	"Direct 256",	"Buffered 256 (most compatible)" };


// Sound
struct SoundStruct {
	UInt32 sampleOffset, samplesQueued;
	UInt16 *soundBuf; //[SOUND_BUFSIZE]; // Ringbuffer for sound
	UInt16 *nextWritePos;
	Boolean makeNoise;
	UInt32 noiseOffs;
} soundStruct = {0, 0, 0, NULL, false, 0};
SndStreamRef sndStream = NULL;

// Handle notification about system goes to sleep
Boolean g_bSystemOffRequested = false;


//***********************************************************
// Procedure: PilotMain
//
// Parameter: cmd           
//            cmdPBP
//            launchFlags
//
// Returnval: -
//
// Description: Startup routine. 
//***********************************************************
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	UInt32 error=0;
  SysAppLaunchCmdCardType *pCmdCard;
  
	switch (cmd) 
	{
		case sysAppLaunchCmdNormalLaunch:
			CheckCrash(); // Check, then set "crashed" flag in prc ('CRSH' 0)
			error = Launch();
			NoCrash(); // Clear "crashed" flag
			break;
		case sysAppLaunchCmdCardLaunch:
		  pCmdCard = (SysAppLaunchCmdCardType *) cmdPBP;
		  pCmdCard->err = errNone;
      
      bLaunchedFromSD = true;
      
      lastCard = pCmdCard->volRefNum; // Get card id to load/save snapshots
			CheckCrash(); // Check, then set "crashed" flag in prc ('CRSH' 0)

  	  //-----------------------------------------------
  	  // Initialize Frodo.prc on SD card to read resources
  	  //-----------------------------------------------
  	  error = VFSFileOpen(lastCard, pCmdCard->path, vfsModeRead, &mainFileRef);
  	  if(error)
  	  {
    		ShowError("Failed to open Frodo.prc on SD card");
    		return 1;
  	  }

      error = Launch();
      
  	  VFSFileClose(mainFileRef);
      
			NoCrash(); // Clear "crashed" flag
  	  break;

		case sysAppLaunchCmdFind:
			break;
	}

	return error;
}


//***********************************************************
// Procedure: CheckCrash
//
// Parameter: -
//
// Returnval: -
//
// Description: Check if Frodo crashed last time it was run.
//***********************************************************
void CheckCrash(void)
{
	UInt16 iDummy;
	Int16 iResult;
  UInt8 bData;
  
  iDummy = sizeof(bData);
  iResult = PrefGetAppPreferences(appSignature, IDX_CRASH, &bData, &iDummy, false);
  if(iResult != noPreferenceFound)
    crashed = bData;
  else
    crashed = true;
  bData = 1;
  PrefSetAppPreferences(appSignature, IDX_CRASH, 1, &bData, sizeof(bData), false);
}


//***********************************************************
// Procedure: NoCrash
//
// Parameter: -
//
// Returnval: -
//
// Description: We made it out alive -> Set flag for next start
//              of Frodo.
//***********************************************************
void NoCrash(void)
{
  UInt8 bData = 0;
  PrefSetAppPreferences(appSignature, IDX_CRASH, 1, &bData, sizeof(bData), false);
}


//***********************************************************
// Procedure: NotifySystemOffRequest
//
// Parameter: pNotifyParam    Parameter of notification           
//
// Returnval: Errorcode
//
// Description: Callback for notification if user switch power off. 
//***********************************************************
Err NotifySystemOffRequest(SysNotifyParamType *pNotifyParam)
{
  Boolean *pSysOffRequest;
  SleepEventParamType *pSleepParam = (SleepEventParamType *) pNotifyParam->notifyDetailsP;
  if(pSleepParam->reason == sysSleepPowerButton)
  {
    // User used power off
    pSysOffRequest = (Boolean *) pNotifyParam->userDataP;
    if(!*pSysOffRequest)
    {
      *pSysOffRequest = true;
      pSleepParam->deferSleep++;
    }
  }
  return errNone;
}


//***********************************************************
// Procedure: GetResource
//
// Parameter: type    Resource type identifier
//            resID   Resource ID
//
// Returnval: MemHandle or 0 on error
//
// Description: Load resource from internal db or SD card
//***********************************************************
MemHandle GetResource(DmResType type, DmResID resID)
{
  MemHandle memH = 0;
  
  if(bLaunchedFromSD)
  {
    if(VFSFileDBGetResource(mainFileRef, type, resID, &memH))
      memH = 0;
  }
  else
  {
    memH = DmGetResource(type, resID);
  }
  
  return memH;
}


//***********************************************************
// Procedure: FreeResource
//
// Parameter: resHandle   Loaded resource handle
//
// Returnval: -
//
// Description: Free resource.
//***********************************************************
void FreeResource(MemHandle resHandle)
{
  if(resHandle)
  {
    if(bLaunchedFromSD)
      MemHandleFree(resHandle);
    else
      DmReleaseResource(resHandle);
  }
}


//***********************************************************
// Procedure: Launch
//
// Parameter: -
//
// Returnval: Errorcode or 0 if successful
//
// Description: Start emulation.
//***********************************************************
UInt32 Launch()
{
	UInt32 error=0;
	
	// Structured programming 8-/
	if (!(error=CheckPrerequisites())) 
	{
	  // Get key mask for future restoration before we change it when loading preferences
	  lKeyMask = KeySetMask(0);
	  KeySetMask(lKeyMask);
	  if (!(error=CreateArmlet())) 
	  {
	    if (!(error=CreateArmletData())) 
	    {
	      InitDB();
				if (!(error=InitGraphicsSound())) 
				{
				  // Register notification for power off
			    error = SysNotifyRegister(0, appSignature, sysNotifySleepRequestEvent, NotifySystemOffRequest, 
			      sysNotifyNormalPriority, &g_bSystemOffRequested);
				  
				  // Clear send keys buffer
				  MemSet(strSendKeysBuffer, SENDKEYSBUFFERSIZE, 0);
				  pSendKeysBufferPos = strSendKeysBuffer;
				  
				  error = RunArmlet();
				  
				  // Unregister notification for power off
				  SysNotifyUnregister(0, appSignature, sysNotifySleepRequestEvent, sysNotifyNormalPriority);
				  
				  // Restore key mask
				  KeySetMask(lKeyMask);
				  
				  if (armletData->cmd & CMD68K_SAVE)
				  {
				    // With it's dying breath, the armlet gave us a snapshot
				    if(bLaunchedFromSD)
				      SaveSnapshot2SD(DEFAULT_SNAP_NAME, lastCard);
				    else
  				    SaveSnapshot(DEFAULT_SNAP_NAME, 0);
          }
          
					if (prefsPalm)
					{
						// save last snapshot name and index to prefs
						prefsPalm->lastSnapshotIndex = lastIndex;
						StrNCopy(prefsPalm->lastSnapshotName, lastName, 31);
						// last chance to save changed palm settings
						SavePrefsPalm();				  
			  	}
				  switch (swap32(error)) 
				  {
						case ERR_NORMAL_EXIT:
							error=0; 
							break;
						case ERR_NEW_FRODO:
							ShowErrorResStr(StringErrOutOfMemoryCreating, "Frodo"); 
							break;
						case ERR_NEW_C64:
							ShowErrorResStr(StringErrOutOfMemoryCreating, "C64"); 
							break;
						case ERR_NEW_CPU:
							ShowErrorResStr(StringErrOutOfMemoryCreating, "CPU"); 
							break;
						case ERR_NEW_JOB1541:
							ShowErrorResStr(StringErrOutOfMemoryCreating, "1541Job"); 
							break;	
						case ERR_NEW_CPU1541:
							ShowErrorResStr(StringErrOutOfMemoryCreating, "1541CPU"); 
							break;
						case ERR_NEW_VIC:
							ShowErrorResStr(StringErrOutOfMemoryCreating, "VIC"); 
							break;
						case ERR_NEW_SID:
							ShowErrorResStr(StringErrOutOfMemoryCreating, "SID"); 
							break;
						case ERR_NEW_CIA1:
							ShowErrorResStr(StringErrOutOfMemoryCreating, "CIA1"); 
							break;
						case ERR_NEW_CIA2:
							ShowErrorResStr(StringErrOutOfMemoryCreating, "CIA2"); 
							break;
						case ERR_NEW_IEC:
							ShowErrorResStr(StringErrOutOfMemoryCreating, "IEC"); 
							break;
						default:
							ShowErrorResLong(StringErrUnknownError, error);
				  }
				  FreeGraphicsSound();
				} // Init GraphicsSound
				FreeDB();
	      FreeArmletData();
	    } // Create ArmletData
	    else 
	    { // ArmletData
	      ShowErrorResLong(StringErrCreateArmletData, error);
	    }
	    FtrPtrFree(appSignature, 0); // Remove the armlet feature
	  } // Create Armlet
	}  // Check Prerequisites
	
  if(g_bSystemOffRequested)
  {
    // Quit because of power off -> now OK to power off
    EvtEnqueueKey(resumeSleepChr, 0, commandKeyMask);
    g_bSystemOffRequested = false;
  }
  
	return error;
}


//***********************************************************
// Procedure: RunArmlet
//
// Parameter: -
//
// Returnval: Errorcode or 0 if successful
//
// Description: Finally, go ARM!
//***********************************************************
UInt32 RunArmlet()
{
	UInt32 error;
	
	FrmGotoForm(MainForm);
	error = PceNativeCall((MemPtr) armlet, armletData);
	return armletData->errCode;
}


//***********************************************************
// Procedure: SetKeyMask
//
// Parameter: -
//
// Returnval: -
//
// Description: Set correct key mask.
//***********************************************************
void SetKeyMask(void)
{
	UInt32 mask=0;
	int lKey;
	
	if (!prefsPalm) 
		return;
	// compute mask for all mapped/used keys	
	mask = prefsPalm->KeyUp | prefsPalm->KeyDown | prefsPalm->KeyLeft | prefsPalm->KeyRight | prefsPalm->KeyFire;
	for(lKey = 0; lKey < 66; ++lKey)
		mask |= prefsPalm->KeyMapping[lKey];
	usedKeys = mask;	
	KeySetMask(~mask);
}


//***********************************************************
// Procedure: SetPaused
//
// Parameter: b   Flag if pause (true) or resume (false)
//
// Returnval: -
//
// Description: Pause/resume
//***********************************************************
void SetPaused(Boolean b)
{
  PauseSound(b);
	pause = b;

	// update statistics if paused because RepaintStatistics is called from VBlank only when not paused
	if (prefsPalm->paused || pause)
	{
		WinHandle w = WinGetDrawWindow(); // save current draw window, can be anything
		WinHandle m = (WinHandle) FrmGetFormPtr(MainForm);
		if (m) 
		{
			WinSetDrawWindow(m); 
			RepaintStatistics(0); // update pause state on screen
			if (w) 
				WinSetDrawWindow(w); // restore previous draw window
		}
	}
}


//***********************************************************
// Procedure: StringForKey
//
// Parameter: k   Keycode
//
// Returnval: Name of key
//
// Description: Get name of key of Palm device
//***********************************************************
char *StringForKey(UInt32 k)
{
	UInt32 i, j;
	
	for (i=0, j=1; i<32; i++, j = j<<1)
		if (k&j) break;
	
	return keys[i];
}


//***********************************************************
// Procedure: CheckJoyPen
//
// Parameter: -
//
// Returnval: -
//
// Description: Joystick emulation using pen.
//***********************************************************
void CheckJoyPen(void)
{
	if (prefsPalm->JoyPen & JOYPEN_SCREEN) 
	{
		// Main screen
		armletData->joystick |= CheckJoyPen2(0,0,160,136);
	}
	if (prefsPalm->JoyPen & JOYPEN_GRAFFITI) 
	{
		// Bottom right (on the joystick indicator)
		if(prefs->Joystick1On)
  		armletData->joystick |= CheckJoyPen2(135, 139, 10,10);
  	else
  		armletData->joystick |= CheckJoyPen2(147, 139, 10,10);
	}
}


//***********************************************************
// Procedure: CheckJoyPen2
//
// Parameter: x   Start position of area
//            y   Start position of area
//            sz  Extend of area
//
// Returnval: Direction of joystick
//
// Description: The JoyPen joystick emulation uses the pen to 
//              specify joystick direction.
//              Basically, you have an origin and and extent, 
//              then see which octant the pen is in.
//***********************************************************
UInt32 CheckJoyPen2(int32 x, int32 y, int32 szx, int32 szy)
{
	Int32 dx = penX - x - (szx>>1); 
	Int32 dy = penY - y - (szy>>1);
	Int32 ax = dx;
	Int32 ay = dy;

	// In the zone?
	if (!penDown || penX < x || penY < y || penX >= x+szx || penY >= y+szy)
		return 0;

	// ax, ay = abs(dx), abs(dy)
	if (ax<0)
		ax = -ax;
	if (ay<0)
		ay = -ay;

	// tan v=sin v/cos v = y/x
	// tan pi/8 = y/x = 0.4142
	// x*4142 = y*10000
	// Ie. in the below means "In the two left-right octants"
	if (ax*4142 > ay*10000) 
	{
		// Left, right
		if (dx < 0)
			return swap32(JOYLEFT);
		else
			return swap32(JOYRIGHT);
	} 
	else if (ay*4142 > ax*10000) 
	{
		// Up, down
		if (dy < 0)
			return swap32(JOYUP);
		else
			return swap32(JOYDOWN);
	} 
	else if (dx < 0) 
	{
		// UpLeft, DownLeft
		if (dy < 0)
			return swap32(JOYUP | JOYLEFT);
		else
			return swap32(JOYDOWN | JOYLEFT);
	} 
	else 
	{
		// UpRight, DownRight
		if (dy<0)
			return swap32(JOYUP | JOYRIGHT);
		else
			return swap32(JOYDOWN | JOYRIGHT);
	}
}


//***********************************************************
// Procedure: CheckJoyKey
//
// Parameter: -
//
// Returnval: -
//
// Description: Joystick keys on the keyboard
//              Called after CheckJoyPen, so armletData->joystick
//              may not be 0.
//***********************************************************
void CheckJoyKey(void)
{
	UInt32 j;
	UInt32 k;
	UInt32 j2;

	j2 = j = swap32(armletData->joystick);
	k = KeyCurrentState();
	if (k & prefsPalm->KeyUp) 
		j |= JOYUP;
	if (k & prefsPalm->KeyLeft) 
		j |= JOYLEFT;
	if (k & prefsPalm->KeyRight) 
		j |= JOYRIGHT;
	if (k & prefsPalm->KeyDown) 
		j |= JOYDOWN;
	if (k & prefsPalm->KeyFire) 
		j |= JOYFIRE;

	if (j != j2)
		keyStateChanged=true;
	armletData->joystick = swap32(j);
}


//***********************************************************
// Procedure: CheckKeyKey
//
// Parameter: -
//
// Returnval: -
//
// Description: Map Palm keys to C64 keys
//***********************************************************
void CheckKeyKey(void)
{
	UInt32 k;
  UInt8 key;
  UInt32 i;
  
	k = KeyCurrentState();

  for(i=0; i<66; ++i)
  {
    // Check if key pressed
    if(k & prefsPalm->KeyMapping[i] && !(g_lLastKeyState & prefsPalm->KeyMapping[i]))
    {
      key = i;
	  	keyStateChanged = true;
      armletData->key = swap32((KEYDOWN<<16) | ((key_byte[key&0x7f]<<8) | ((key&128)?0x800:0)) | key_bit[key&0x7f]);
      g_lLastKeyState = g_lLastKeyState | prefsPalm->KeyMapping[i];
      break; // Only on state-change per frame -> exit i-loop
    }
    // Check if key was pressed before and now released
    if(g_lLastKeyState & prefsPalm->KeyMapping[i] && !(k & prefsPalm->KeyMapping[i]))
    {
      key = i;
	  	keyStateChanged = true;
      armletData->key = swap32((KEYUP<<16) | ((key_byte[key&0x7f]<<8) | ((key&128)?0x800:0)) | key_bit[key&0x7f]);
      g_lLastKeyState = g_lLastKeyState & ~(prefsPalm->KeyMapping[i]);
      break; // Only on state-change per frame -> exit i-loop
    }
  }
}


//*******************************************************************************************
//*******************************************************************************************
// Dialogs and forms
//*******************************************************************************************
//*******************************************************************************************

//***********************************************************
// Procedure: ShowDialog
//
// Parameter: id    ID of dialog to display
//
// Returnval: ID of selected button
//
// Description: Display dialog and wait for end of dialog
//              Show one of the dialogs. Dialogs may or may not 
//              have a Init function, an Exit function and an event handler.
//              Init is called after the form is created, but before it is displayed.
//              Exit is called after the form is closed, but before it is destroyed.
//***********************************************************
UInt16 ShowDialog(UInt16 id)
{
	FormType *f = FrmInitForm(id);
	UInt16 choice = 0;
	void (*initP)(FormType *) = NULL;
	void (*exitP)(FormType *, UInt16) = NULL;
  UInt32 mask = KeySetMask(lKeyMask); // Enable all keys

	switch (id) 
	{
		case GfxSoundForm:
			FrmSetEventHandler(f, HandleEventGfxSound);
			initP = InitGfxSoundDialog;
			exitP = ExitGfxSoundDialog;
			break;
			
		case JoystickForm:
			FrmSetEventHandler(f, HandleEventJoystick);
			initP = InitJoystickDialog;
			break;
			
		case ObscureForm:
			initP = InitObscureDialog;
			exitP = ExitObscureDialog;
			break;
			
		case KeyboardForm:
			FrmSetEventHandler(f, HandleEventKeyboard);
			initP = InitKeyboardDialog;
			break;
			
		case SampleKeyForm:
			FrmSetEventHandler(f, HandleEventSampleKey);
			initP = InitSampleKeyDialog;
			break;

		case ColorsForm:
			FrmSetEventHandler(f, HandleEventColors);
			initP = InitColorsDialog;
			exitP = ExitColorsDialog;
			break;

		case CommandsForm:
			FrmSetEventHandler(f, HandleEventCommands);
			initP = InitCommandsDialog;
			exitP = ExitCommandsDialog;
			break;
	}
	
	if (initP) 
		initP(f);
	choice = FrmDoDialog(f);
	if (exitP) 
		exitP(f, choice);
	FrmDeleteForm(f);

  // Back to normal.
  KeySetMask(mask);

	return choice;
}


//***********************************************************
// Procedure: ShowSaveAsDialog
//
// Parameter: s   Default file name
//
// Returnval: New filename
//
// Description: Show the "Save as" requester, with an optional 
//              default file name
//***********************************************************
char *ShowSaveAsDialog(char *s)
{
	FormType *f = FrmInitForm(SaveAsForm);
	UInt16 choice = 0;
	static char b[32];
	const char *t;
	char *retval = NULL;
	UInt16 card;
	UInt16 index;
	ControlType *pCheckbox;
	
	if (s)
		SetText(f, SaveAsNameField, s);
	FrmSetFocus(f, FrmGetObjectIndex(f, SaveAsNameField));
	
	card = SearchValidCard();
	if(card == 0xffff)
	{
	  // Disable checkbox for SD-Card (with checkbox not selected)
    index = FrmGetObjectIndex(f, SaveAsSDCardCheckbox);
    pCheckbox = (ControlType *) FrmGetObjectPtr(f, index);
	  CtlSetEnabled(pCheckbox, false);
	}
  else if(lastCard != 0xffff)
  {
    // Mark checkbox as selected
    SetValue(f, SaveAsSDCardCheckbox, true);

    if(bLaunchedFromSD)
    {
  	  // Launched from SD -> snapshots can only be saved on SD card
  	  // Disable checkbox for SD-Card (with checkbox SELECTED!)
      index = FrmGetObjectIndex(f, SaveAsSDCardCheckbox);
      pCheckbox = (ControlType *) FrmGetObjectPtr(f, index);
  	  CtlSetEnabled(pCheckbox, false);
    }
  }
  	
	choice = FrmDoDialog(f);
	if (choice == SaveAsOKButton) 
	{
		t = GetText(f, SaveAsNameField);
		if (t != NULL) 
		{
			StrNCopy(b, t, 31);
			b[31] = '\0';
			retval = b;
		}
    if(GetValue(f, SaveAsSDCardCheckbox))
      lastCard = card;
    else
      lastCard = 0xffff;
	}
	FrmDeleteForm(f);
	return retval;
}


//***********************************************************
// Procedure: SampleKey
//
// Parameter: org   Previous used key
//
// Returnval: Selected key
//
// Description: User selects key on Palm device.
//***********************************************************
UInt32 SampleKey(UInt32 org)
{
	UInt32 choice;

	// The eventHandler sets sampledKey everytime the 'Sample' button is pressed.
	sampledKey = 0;
	choice = ShowDialog(SampleKeyForm);
	
	// Back to normal.
	KeySetMask(lKeyMask);
	if (choice == SampleKeyOKButton)
		return sampledKey;
	else
		return org;
}


//***********************************************************
// Procedure: InitColorsDialog
//
// Parameter: f   Form of dialog
//
// Returnval: -
//
// Description: Initialize dialog to set colors.
//***********************************************************
void InitColorsDialog(FormType *f)
{
	UInt16 lIndex;
  TableType *pTable;
	UInt16 idx;
	ListType *l;
	UInt32 pres;
  
	idx = FrmGetObjectIndex(f, ColorsPresetList);
	if (idx != frmInvalidObjectId) 
	{
		l = (ListType *) FrmGetObjectPtr(f, idx);
		LstSetListChoices(l, ColorPresList, NCOLORPRESET);
		pres = 0;
		if(pres > NCOLORPRESET-1) 
			pres = 0;
		LstSetSelection(l, pres);
		FrmCopyLabel(f, ColorsPresetPopTrigger, ColorPresList[pres]);
	}
  
	lIndex = FrmGetObjectIndex(f, ColorsTable);
  pTable = (TableType *) FrmGetObjectPtr(f, lIndex);
  TblSetItemStyle(pTable, 0, 0, customTableItem);
  TblSetRowHeight(pTable, 0, 20);
  TblSetColumnUsable(pTable, 0, true);
  TblSetCustomDrawProcedure(pTable, 0, ColorsTableDraw);

  // Set new color palette
  WinPalette(winPaletteSet, 2, 16, (RGBColorType *) prefsPalm->colorTable);
}


//***********************************************************
// Procedure: ExitColorsDialog
//
// Parameter: f       Form if obscure dialog
//            choice  Flag if user selected OK button
//
// Returnval: -
//
// Description: Get settings from colors dialog
//***********************************************************
void ExitColorsDialog(FormType *f, UInt16 choice)
{
	if (choice != ColorsOKButton)
	{
		// Reset current used palette
		WinPalette(winPaletteSet, 0, 16, (RGBColorType *) prefsPalm->colorTable);
    return;
	}
	
	// Get current modified palette
	WinPalette(winPaletteGet, 2, 16, (RGBColorType *) prefsPalm->colorTable);
	SavePrefsPalm();
	WinPalette(winPaletteSet, 0, 16, (RGBColorType *) prefsPalm->colorTable);
}


//***********************************************************
// Procedure: InitKeyboardDialog
//
// Parameter: f   Form of dialog
//
// Returnval: -
//
// Description: Initialize dialog to set preferences for keyboard.
//***********************************************************
void InitKeyboardDialog(FormType *f)
{
	UInt16 lIndex;
  TableType *pTable;
  
	lIndex = FrmGetObjectIndex(f, KeyboardMapTable);
  pTable = (TableType *) FrmGetObjectPtr(f, lIndex);
  TblSetItemStyle(pTable, 0, 0, customTableItem);
  TblSetRowHeight(pTable, 0, 121);
  TblSetColumnUsable(pTable, 0, true);
  TblSetCustomDrawProcedure(pTable, 0, KeyboardTableDraw);
}


//***********************************************************
// Procedure: InitGfxSoundDialog
//
// Parameter: f   Form to display dialog
//
// Returnval: -
//
// Description: Init dialog to set graphics and sound options
//***********************************************************
void InitGfxSoundDialog(FormPtr f)
{
	UInt16 idx;
	ListType *l;
	UInt32 fs, sm;
  char msg[16];
  
	idx = FrmGetObjectIndex(f, GfxSoundFrameskipList);
	if (idx != frmInvalidObjectId) 
	{
		l=(ListType *) FrmGetObjectPtr(f, idx);
		LstSetListChoices(l, frameskipList, NSKIPLIST);
		if (prefs->AdaptiveFrameSkip)
			fs = 0;
		else
			fs = swap32(prefs->SkipFrames);
		if (fs > NSKIPLIST-1)
		  fs = NSKIPLIST-1;
		LstSetSelection(l, fs);
		FrmCopyLabel(f, GfxSoundFrameskipPopTrigger, frameskipList[fs]);
	}
	
	idx = FrmGetObjectIndex(f, GfxSoundScreenModeList);
	if (idx != frmInvalidObjectId) 
	{
		l = (ListType *) FrmGetObjectPtr(f, idx);
		sm = prefsPalm->ScreenMode;
		LstSetListChoices(l, screenmodeList, NSCREENMODE);
		if (sm < 0 || sm >= NSCREENMODE) 
		{
			ShowErrorResLong(StringErrBadScreenMode, sm);
			sm = 0;
		}
		LstSetSelection(l, sm);
		FrmCopyLabel(f, GfxSoundScreenModePopTrigger, screenmodeList[sm]);
	}
	
	SetValue(f, GfxSoundSIDFiltersCheckbox, prefs->SIDFilters ? 1:0);
	SetValue(f, GfxSoundTurboCheckbox, prefs->LimitSpeed ? 0:1);
	SetValue(f, GfxSoundVolumeFeedbackSliderControl, prefsPalm->Volume);
	if (!prefsPalm->Volume)
		StrPrintF(msg, "Vol: Mute");
	else
		StrPrintF(msg, "Vol: %ld%%", (100 * prefsPalm->Volume)>>10);
	SetText(f, GfxSoundVolumeLabel, msg);

	SetValue(f, GfxSoundSpeedLimitFeedbackSliderControl, prefsPalm->SpeedLimit);
	StrPrintF(msg, "Speed: %d%%", prefsPalm->SpeedLimit);
	SetText(f, GfxSoundSpeedLimitLabel, msg);
}


//***********************************************************
// Procedure: ExitGfxSoundDialog
//
// Parameter: f       Form of dialog
//            choice  Flag if user selected OK
//
// Returnval: -
//
// Description: Get settings from dialog of graphics and
//              sound preferences.
//***********************************************************
void ExitGfxSoundDialog(FormPtr f, uint16 choice)
{
	UInt16 idx;
	ListType *l;
	UInt32 v;

	MakeNoise(false, 0);
	if (choice != GfxSoundOKButton) 
		return;

	prefs->SIDFilters = GetValue(f, GfxSoundSIDFiltersCheckbox);
	prefs->LimitSpeed = !GetValue(f, GfxSoundTurboCheckbox);
	prefsPalm->SpeedLimit = GetValue(f, GfxSoundSpeedLimitFeedbackSliderControl);
	v = GetValue(f, GfxSoundVolumeFeedbackSliderControl);
	prefsPalm->Volume = v;
	
	idx = FrmGetObjectIndex(f, GfxSoundFrameskipList);
	if (idx != frmInvalidObjectId) 
	{
		l = (ListType *) FrmGetObjectPtr(f, idx);
		idx = LstGetSelection(l);
		if (idx == 0)
			prefs->AdaptiveFrameSkip = 1;
		else if (idx > 0 && idx <= NSKIPLIST-1) 
		{
			prefs->AdaptiveFrameSkip = 0;
			prefs->SkipFrames = swap32(idx);
		}
	}

	idx = FrmGetObjectIndex(f, GfxSoundScreenModeList);
	if (idx != frmInvalidObjectId) 
	{
		l = (ListType *) FrmGetObjectPtr(f, idx);
		idx = LstGetSelection(l);
		ChangeScreenMode(idx, false, false);
		if (FrmGetActiveForm()) 
			FrmDrawForm(FrmGetActiveForm());
	}

	if (v) 
	{
		canSound = true;
		StartSound();
		prefs->SIDOn = 1;
	} 
	else
	{
		StopSound();
    prefs->SIDOn = 0;
  }
  
	SavePrefs();
  SavePrefsPalm();
	armletData->cmd |= CMD68K_SETPREFS;
}


//***********************************************************
// Procedure: InitJoystickDialog
//
// Parameter: f   Form of dialog
//
// Returnval: -
//
// Description: Initialize dialog to set preferences for joystick.
//***********************************************************
void InitJoystickDialog(FormType *f)
{
	SetText(f, JoystickUpLabel, StringForKey(prefsPalm->KeyUp));
	SetText(f, JoystickLeftLabel, StringForKey(prefsPalm->KeyLeft));
	SetText(f, JoystickRightLabel, StringForKey(prefsPalm->KeyRight));
	SetText(f, JoystickDownLabel, StringForKey(prefsPalm->KeyDown));
	SetText(f, JoystickFireLabel, StringForKey(prefsPalm->KeyFire));
	SetValue(f, JoystickEmulate1Checkbox, prefs->Joystick1On? 1:0);
	SetValue(f, JoystickEmulate2Checkbox, prefs->Joystick2On? 1:0);
	SetValue(f, JoystickJoyPenScreenCheckbox, (prefsPalm->JoyPen & JOYPEN_SCREEN)? 1:0);
	SetValue(f, JoystickJoyPenGraffitiCheckbox, (prefsPalm->JoyPen & JOYPEN_GRAFFITI)? 1:0);
	dlgUp = prefsPalm->KeyUp;
	dlgLeft = prefsPalm->KeyLeft;
	dlgRight = prefsPalm->KeyRight;
	dlgDown = prefsPalm->KeyDown;
	dlgFire = prefsPalm->KeyFire;
}


//***********************************************************
// Procedure: InitSampleKeyDialog
//
// Parameter: f   Form to display dialog
//
// Returnval: -
//
// Description: Initialize dialog to sample keys
//***********************************************************
void InitSampleKeyDialog(FormType *f)
{
  // Lock all keys
  KeySetMask(0);
}


//***********************************************************
// Procedure: InitObscureDialog
//
// Parameter: f   Form to display dialog
//
// Returnval: -
//
// Description: Edit preferences for cycles/frame, sprite collisions...
//***********************************************************
void InitObscureDialog(FormType *f)
{
	UInt16 idx;
	ListType *l;
	UInt32 currentmode;

  // Fill list with options for saving to SD card
	idx = FrmGetObjectIndex(f, ObscureSaveToSDList);
	if (idx != frmInvalidObjectId)
	{
		l = (ListType *) FrmGetObjectPtr(f, idx);
		currentmode = prefsPalm->SaveToSDCard;
		LstSetListChoices(l, save2SDList, NSAVE2SD_MODE);
		if (currentmode < 0 || currentmode >= NSAVE2SD_MODE)
			currentmode = 0;
		LstSetSelection(l, currentmode);
		FrmCopyLabel(f, ObscureSaveToSDPopTrigger, save2SDList[currentmode]);
	}

	SetValue(f, ObscureSingleCycleCheckbox, prefs->SingleCycleEmulation ? 1:0);
	SetValue(f, ObscureSpritesCheckbox, prefs->SpritesOn ? 1:0);
	SetValue(f, ObscureSpriteCollisionsCheckbox, prefs->SpriteCollisions ? 1:0);
	SetValue(f, ObscureFastResetCheckbox, prefs->FastReset ? 1:0);
	SetValue(f, ObscureCIAHackCheckbox, prefs->CIAIRQHack ? 1:0);
	SetValue(f, ObscureEmulate1541Checkbox, prefs->Emul1541Proc ? 1:0);
	SetValue(f, ObscureTrackNumberCheckbox, prefsPalm->DisplayTrackNum ? 1:0);
}


//***********************************************************
// Procedure: ExitObscureDialog
//
// Parameter: f       Form if obscure dialog
//            choice  Flag if user selected OK button
//
// Returnval: -
//
// Description: Get settings from obscure dialog
//***********************************************************
void ExitObscureDialog(FormType *f, UInt16 choice)
{
	UInt16 idx;
	ListType *l;
	
	if (choice != ObscureOKButton) 
		return;
	
  prefs->SingleCycleEmulation = GetValue(f, ObscureSingleCycleCheckbox);
	prefs->SpritesOn = GetValue(f, ObscureSpritesCheckbox);
	prefs->SpriteCollisions = GetValue(f, ObscureSpriteCollisionsCheckbox);
	prefs->FastReset = GetValue(f, ObscureFastResetCheckbox);
	prefs->CIAIRQHack = GetValue(f, ObscureCIAHackCheckbox);
	prefs->Emul1541Proc = GetValue(f, ObscureEmulate1541Checkbox);
	prefsPalm->DisplayTrackNum = GetValue(f, ObscureTrackNumberCheckbox);

  // Get settings for saving to sd card
	idx = FrmGetObjectIndex(f, ObscureSaveToSDList);
	if (idx != frmInvalidObjectId)
	{
		l = (ListType *) FrmGetObjectPtr(f, idx);
		idx = LstGetSelection(l);
		prefsPalm->SaveToSDCard = idx;
	}

	if(prefs->Emul1541Proc && prefs->DriveType == swap32(DRVTYPE_T64))
	{
	  // No emulation if t64 or p00 file
	  prefs->Emul1541Proc = 0;
	  ShowInfoRes(StringWarn1541Off);
	}
	
	SavePrefs();
  SavePrefsPalm();
	armletData->cmd |= CMD68K_SETPREFS;
}


//***********************************************************
// Procedure: ColorsTableDraw
//
// Parameter: pTable  Pointer to table to draw
//            row     Row of current cell
//            col     Column of current cell
//            bounds  Area of current cell
//
// Returnval: -
//
// Description: Draw content of a cell
//***********************************************************
void ColorsTableDraw(void *pTable, Int16 row, Int16 col, RectangleType *bounds)
{
  UInt16 i;
  RGBColorType oldColor;
  RectangleType rect;
  
  // Get current color for drawing
  WinSetForeColorRGB(NULL, &oldColor);
  
  // Draw rect for each color
  for(i=0; i<16; ++i)
  {
    WinSetForeColor(i+2);
    rect.topLeft.x = bounds->topLeft.x + (i * 10) % 80;
    if(i < 8)
      rect.topLeft.y = bounds->topLeft.y;
    else
      rect.topLeft.y = bounds->topLeft.y + 10;
    rect.extent.x = 10;
    rect.extent.y = 10;
    WinDrawRectangle(&rect, 0);
  }

  // Restore color for drawing
  WinSetForeColorRGB(&oldColor, NULL);
}


//***********************************************************
// Procedure: KeyboardTableDrawItem
//
// Parameter: lIndex  Index of key
//
// Returnval: -
//
// Description: Draw content of a cell
//***********************************************************
void KeyboardTableDrawItem(UInt32 lIndex)
{
  Int16 lColumn;
  Int16 x, y;
  char *strKey;
  
  lColumn = C64KeyToDisplay[lIndex] / 17;
  x = g_lColPos[lColumn * 2 + 1];
  y = 16 + (C64KeyToDisplay[lIndex] % 17) * 7;
  strKey = StringForKey(prefsPalm->KeyMapping[lIndex]);
  WinDrawChars(strKey, StrLen(strKey), x, y);
}


//***********************************************************
// Procedure: KeyboardTableDraw
//
// Parameter: pTable  Pointer to table to draw
//            row     Row of current cell
//            col     Column of current cell
//            bounds  Area of current cell
//
// Returnval: -
//
// Description: Draw content of a cell
//***********************************************************
void KeyboardTableDraw(void *pTable, Int16 row, Int16 col, RectangleType *bounds)
{
  UInt16 i;
  MemHandle bmpHdl = 0;
  BitmapPtr bmpPtr;

	bmpHdl = GetResource(bitmapType, EditKeyMappingBitmap);
	if(bmpHdl)
	{
		bmpPtr = MemHandleLock(bmpHdl);
		if(bmpPtr)
		{
      WinDrawBitmap(bmpPtr, 2, 14);
	  	MemHandleUnlock(bmpHdl);
		}
		FreeResource(bmpHdl);
  }

  FntSetFont(fntStandardID);
  for(i=0; i<66; ++i)
    KeyboardTableDrawItem(i);
}


//***********************************************************
// Procedure: SetValue
//
// Parameter: f   Form with control
//            id  ID of control
//            v   New value for control
//
// Returnval: True if new value set.
//
// Description: Set the value of a control.
//***********************************************************
Boolean SetValue(FormType *f, UInt16 id, UInt16 v)
{
	UInt16 idx = FrmGetObjectIndex(f, id);
	void *c;

	if (idx == frmInvalidObjectId)
		return false;

	c = FrmGetObjectPtr(f, idx);
	CtlSetValue((ControlType *) c, v);

	return true;
}


//***********************************************************
// Procedure: GetValue
//
// Parameter: f   Form with control
//            id  ID of control
//
// Returnval: Value of a control as integer
//
// Description: Get the value of a control
//***********************************************************
UInt16 GetValue(FormType *f, UInt16 id)
{
	UInt16 idx = FrmGetObjectIndex(f, id);
	void *c;

	if (idx == frmInvalidObjectId)
		return false;

	c = FrmGetObjectPtr(f, idx);
	return CtlGetValue((ControlType *) c);
}


//***********************************************************
// Procedure: SetText
//
// Parameter: f   Form with control
//            id  ID of control
//            s   Text
//
// Returnval: True if text of control changed
//
// Description: Set a controls label or text to a string
//***********************************************************
Boolean SetText(FormType *f, UInt16 id, char *s)
{
	UInt16 idx = FrmGetObjectIndex(f, id);
	FormObjectKind k;
	void *c;
	FieldType *fl;

	if (idx == frmInvalidObjectId)
		return false;

	c = FrmGetObjectPtr(f, idx);
	k = FrmGetObjectType(f, idx);
	switch (k) 
	{
		case frmFieldObj:
			fl = (FieldType *)c;
			FldDelete(fl, 0, FldGetTextLength(fl)-1);
			FldInsert(fl, s, StrLen(s));
			break;
		default:
			FrmCopyLabel(f, id, s);
	}
	return true;
}


//***********************************************************
// Procedure: GetText
//
// Parameter: f   Form with control
//            id  ID of control
//
// Returnval: Text or caption of control
//
// Description: Get a controls label or text
//***********************************************************
const char *GetText(FormType *f, UInt16 id)
{
	UInt16 idx = FrmGetObjectIndex(f, id);
	FormObjectKind k;
	void *c;

	if (idx == frmInvalidObjectId)
		return NULL;

	c = FrmGetObjectPtr(f, idx);
	k = FrmGetObjectType(f, idx);
	switch (k) 
	{
		case frmFieldObj:
			return FldGetTextPtr((FieldType *) c);
		default:
			return CtlGetLabel((ControlType *) c);
	}
}


//*******************************************************************************************
//*******************************************************************************************
// Event handler
//*******************************************************************************************
//*******************************************************************************************

//***********************************************************
// Procedure: AppHandleEvent
//
// Parameter: e   Event
//
// Returnval: True if event handled in this procedure
//
// Description: Application event handler. Handle form init.
//***********************************************************
Boolean AppHandleEvent(EventPtr e)
{
	Boolean handled = false;
	FormType *f = FrmGetActiveForm();

	switch (e->eType) 
	{
	  case frmLoadEvent:
	    switch (e->data.frmLoad.formID) 
	    {
	      case MainForm:
      		f = FrmInitForm(e->data.frmLoad.formID);
      		FrmSetActiveForm(f);
      		FrmSetEventHandler(f, HandleEventMain);
      		handled = true;
      		break;

	      case FileForm:
      		f = FrmInitForm(FileForm);
      		FrmSetActiveForm(f);
      		FrmSetEventHandler(f, HandleEventFile);
      		handled = true;
      		break;
	    }
	    break;
	  case appStopEvent:
	    FrmEraseForm(f);
	    FrmDeleteForm(f);
	    exiting = true;
	    handled = true;
	    break;

    // Check pen position, I don't trust non-pen events.
	  case penDownEvent:
			// open menu when clicking in upper left corner
			if (e->screenX<5 && e->screenY<5 )
			{
				EvtEnqueueKey(menuChr,0,commandKeyMask);
		    handled = true;
		    break;
			}

		// otherwise continue below
	  case penMoveEvent:
	    penDown = true;
	    penX = e->screenX;
	    penY = e->screenY;
	    break;

	  case penUpEvent:
	    penDown = false;
	    break;
	}
	
	return handled;
}


//***********************************************************
// Procedure: HandleEventMain
//
// Parameter: e   Event
//
// Returnval: Flag if event handled in this procedure
//
// Description: Main event handler, handle on screen keyboard, menu.
//***********************************************************
Boolean HandleEventMain(EventPtr e)
{
	Boolean handled = false;
	FormType *f = FrmGetActiveForm();
	RectangleType r;
	UInt16 i;
	UInt8 key;
	UInt32 j;

	switch (e->eType) 
	{
	  case menuEvent:
	    handled = true;
	    switch (e->data.menu.itemID) 
	    {
	      case GameFilemanager:
      		// Give the armlet a chance to perform a snapshot
      		// before displaying form.
	        doFileForm = true;
      		armletData->cmd |= CMD68K_SAVE;
					if (prefsPalm->paused)
					{
						SetPaused(false); 					// unpause
						pausedStateValid	=	false; 	// restore pause state after first frame is drawn
					}
      		break;
	      case GameSaveAs:
      		doSaveAs=true;
      		armletData->cmd |= CMD68K_SAVE;
					if (prefsPalm->paused)
					{
						SetPaused(false); 				// unpause for creating the snapshot
						pausedStateValid=false;
					}
      		break;
	      case GameQuickload:
      		doQuickload=true;
					if (prefsPalm->paused)
					{
						SetPaused(false); 				// unpause for loading and drawing the first frame
						pausedStateValid=false;
					}
      		break;
	      case GameQuicksave:
      		doQuicksave=true;
      		armletData->cmd |= CMD68K_SAVE;
					if (prefsPalm->paused)
					{
						SetPaused(false); 			// unpause for the snapshot
						pausedStateValid=false; // restore pause state after first frame is drawn
					}
      		break;
	      case GameReset:
      		armletData->cmd |= CMD68K_RESET;
					if (prefsPalm->paused)
					{
						prefsPalm->paused = 0;
						SetPaused(false); 			// unpause and let it run, this is reset
					}
					MenuEraseStatus(0);
      		break;
	      case GamePause:
					prefsPalm->paused = !prefsPalm->paused; // toggle prefs
					SetPaused(prefsPalm->paused != 0);			// make it active
					MenuEraseStatus(0);
      		break;
	      case OptionsAboutFrodo:
      		ShowDialog(AboutForm);
      		break;
	      case OptionsGraphicsandsound:
      		ShowDialog(GfxSoundForm);
      		break;
	      case OptionsJoystick:
      		ShowDialog(JoystickForm);
      		break;
	      case OptionsKeyboard:
      		ShowDialog(KeyboardForm);
      		break;
	      case OptionsObscure:
      		ShowDialog(ObscureForm);
      		break;
	      case OptionsColors:
      		ShowDialog(ColorsForm);
      		break;
	      case OptionsSwapjoysticks:
	        j = prefs->Joystick1On;
      		prefs->Joystick1On = prefs->Joystick2On;
      		prefs->Joystick2On = j;
      		SavePrefs();
      		break;
	      case OptionsSwitchEmulMode:
	        if(prefs->SingleCycleEmulation)
	          prefs->SingleCycleEmulation = 0;
	        else
	          prefs->SingleCycleEmulation = 1;
	        SavePrefs();
        	armletData->cmd |= CMD68K_SETPREFS;
	        break;
	      case OptionsSwitch1541Mode:
	        if(prefs->Emul1541Proc)
	          prefs->Emul1541Proc = 0;
	        else
	        {
	          prefs->Emul1541Proc = 1;
          	if(prefs->DriveType == swap32(DRVTYPE_T64))
          	{
          	  // No emulation if t64 or p00 file
          	  prefs->Emul1541Proc = 0;
          	  ShowInfoRes(StringWarn1541Off);
          	}
	        }
	        SavePrefs();
        	armletData->cmd |= CMD68K_SETPREFS;
	        break;
	      case OptionsCommands:
      		ShowDialog(CommandsForm);
      		break;
	      default:
      		handled = false;
	    }

      if(e->data.menu.itemID >= CommandsMenuFirst && e->data.menu.itemID < CommandsMenuFirst + C64COMMANDS_MAX)
        // C64-Command selected
        StrNCopy(strSendKeysBuffer, prefsPalm->C64Command[e->data.menu.itemID - CommandsMenuFirst], C64COMMANDS_MAXLEN);
      break;

	  case frmOpenEvent:
	    FrmDrawForm(f);
	    handled = true;
	    break;

	  case menuOpenEvent:
      // Create item for each command
      for(i=0; i<C64COMMANDS_MAX; ++i)
      {
        if(prefsPalm->C64CommandTitle[i] == 0 || prefsPalm->C64CommandTitle[i][0] == 0)
        {
          // Entry not defined
          if(i == 0)
            // Create <not set> for first entry if none defined
            MenuAddItem(CommandsMenuDummy, CommandsMenuFirst + i, '1' + i, "<not set>");
          break; // Exit for-loop
        }
        MenuAddItem(CommandsMenuFirst - 1 + i, CommandsMenuFirst + i, '1' + i, prefsPalm->C64CommandTitle[i]);
      }
      
      // Hide dummy entry of command menu
      MenuHideItem(CommandsMenuDummy);
      break;

	  case penDownEvent:
	    /* We need some custom event handling for the onscreen keyboard.
	       If we let the ctlEnterEvent past, FrmHandleEvent will go into a loop tracking
	       the pen until it's released, and the emulation will freeze.
	    */
	    for (i=MainFirstKeyGraphicButton; i<=MainSpaceGraphicButton; i++) 
	    {
	      FrmGetObjectBounds(f, FrmGetObjectIndex(f, i), &r);
	      if (RctPtInRectangle(e->screenX, e->screenY, &r))
      		break;
	    }
	    // Keys are numbered right-to-left, top-to-bottom so space is last
	    if (i>MainSpaceGraphicButton)
	      break;
	    
	    inKey = i;
	    // TODO: Invert key
	    if (inKey != MainCommodoreGraphicButton && inKey != MainShiftLeftGraphicButton && inKey != MainShiftRightGraphicButton) 
	    {
	      // On screen keyboard key down
	      armletData->key = swap32((KEYDOWN<<16) | (key_byte[inKey-MainFirstKeyGraphicButton]<<8) |
		      key_bit[inKey-MainFirstKeyGraphicButton]);
	    }
	    handled = true;
	    break;
	    
	  case penUpEvent:
	    if (inKey) 
	    {
	      switch (inKey) 
	      {
      		case MainCommodoreGraphicButton:
		        // Sticky commodore and left shift
      		  cmdLock = !cmdLock;
	          armletData->key = swap32(((cmdLock? KEYDOWN:KEYUP)<<16) |	(key_byte[inKey-MainFirstKeyGraphicButton]<<8) |
        			key_bit[inKey-MainFirstKeyGraphicButton]);
		        if (shiftLock && cmdLock)
      		    lowerCase = !lowerCase;
	     		  UpdateOnScreenKeyboard(f);
      		  break;

      		// case MainShiftLockGraphicButton:
      		case MainShiftLeftGraphicButton:
      		case MainShiftRightGraphicButton:
      		  shiftLock = !shiftLock;
      		  armletData->key=swap32(((shiftLock? KEYDOWN:KEYUP)<<16) | (key_byte[inKey-MainFirstKeyGraphicButton]<<8) |
      			  key_bit[inKey-MainFirstKeyGraphicButton]);
      		  if (shiftLock && cmdLock)
      		    lowerCase = !lowerCase;
      		  UpdateOnScreenKeyboard(f);
      		  break;

      		default:
      		  // On-screen keyboard key up
      		  armletData->key = swap32((KEYUP<<16) | (key_byte[inKey-MainFirstKeyGraphicButton]<<8) |
      			  key_bit[inKey-MainFirstKeyGraphicButton]);
      		  if ((inKey == MainShiftRightGraphicButton) && cmdLock) 
      		  {
      		    lowerCase = !lowerCase;
      		    UpdateOnScreenKeyboard(f);
      		  } 
      		  else 
      		  {
      		    // TODO: Invert key
      		  }
      	}
	      inKey = 0;
	      handled = true;
	    }
	    break;
	  
	  case keyDownEvent:
	    handled = false;
	    
	    // We only care about some ascii characters.
	    if(e->data.keyDown.chr < 0 || e->data.keyDown.chr > 128)
        break;
	    key = C64KeyNumberFromAscii[e->data.keyDown.chr];
	    // character does not exist on the C64 keyboard
	    if(key == 255)
    		break;
	    armletData->key = swap32((KEYDOWN<<16) | ((key_byte[key&0x7f]<<8) | ((key&128)?0x800:0)) | key_bit[key&0x7f]);

	    /* Need to wait some frames / some ticks for the KEYDOWN to be noticed.
	    Notice the time here.*/
	    ticksLastKeyDown = TimGetTicks();

	    /* savedGraffitiKey will be used later on */
	    savedGraffitiKey = swap32((KEYUP<<16) | ((key_byte[key&0x7f]<<8) | ((key&128)?0x800:0)) | key_bit[key&0x7f]);
		    
	    handled = true;
	    break;

	    // Try to pause when menu goes down
	  case winExitEvent:
	    if (e->data.winExit.exitWindow == (WinHandle) FrmGetFormPtr(MainForm))
	    {
    		SetPaused(true);
				// restore system key mask (for menu navigation)
				KeySetMask(lKeyMask);
			}
	    break;

	  case winEnterEvent:
	    if (e->data.winEnter.enterWindow == (WinHandle) FrmGetFormPtr(MainForm) 
	    &&  e->data.winEnter.enterWindow == (WinHandle) FrmGetFirstForm()) 
	    {
    		if (pausedStateValid) // pause or start emulation only when appropriate
    		{ 
					SetPaused((prefsPalm->paused !=0 ));
				}
			// disable mapped keys again
			SetKeyMask();
	    }
	    break;
	}
	
	return handled;
}


//***********************************************************
// Procedure: HandleEventColors
//
// Parameter: e   Event
//
// Returnval: True if event handled in this procedure
//
// Description: Handle event of dialog to set colors
//***********************************************************
Boolean HandleEventColors(EventPtr e)
{
	UInt16 idx;
	RGBColorType currColor;
	
	switch (e->eType) 
	{
	  case popSelectEvent:
	  	switch(e->data.popSelect.controlID)
	  	{
	  		case ColorsPresetPopTrigger:
	  			idx = e->data.popSelect.selection;
	  			if(idx >= 0 && idx < NCOLORPRESET)
	  				WinPalette(winPaletteSet, 2, 16, &(colorTablePreset[16 * idx]));
	  			break;
	  	}
	  	break;
	  	
	  case penDownEvent:
      if(e->screenY >= 40 && e->screenY < 60 && e->screenX >= 23 && e->screenX < 103)
      {
	      idx = (e->screenX - 23) / 10;
 	      if(e->screenY >= 50)
 	      	idx = idx + 8;
 	      WinPalette(winPaletteGet, 2 + idx, 1, &currColor);
 	      if(UIPickColor(NULL, &currColor, UIPickColorStartRGB, "Select color", NULL))
 	      	WinPalette(winPaletteSet, 2 + idx, 1, &currColor);
  	  }
      break;	  
	}

	return false;
}


//***********************************************************
// Procedure: HandleEventFile
//
// Parameter: e   Event
//
// Returnval: True if event handled in this procedure
//
// Description: Event handler for dialog for file selection.
//***********************************************************
Boolean HandleEventFile(EventPtr e)
{
	Boolean handled = true;
	Boolean doExit = false;
	FormPtr f = FrmGetActiveForm();
	ListPtr l = NULL;
	static Int16 listLastIndex = 0;
	UInt16 i;
	char *s;
	
	switch (e->eType) 
	{
	  case frmOpenEvent:
      KeySetMask(lKeyMask); // Enable all keys
	    FrmDrawForm(f);
	    ScanFiles(f);
			if (listLastIndex>0)
			{
				// preselect last used item
				i = FrmGetObjectIndex(f, FileFileList);
				if (i != frmInvalidObjectId)
					l=(ListType *) FrmGetObjectPtr(f, i);
				if (l) 
      		LstSetSelection(l,listLastIndex);
			}
	    break;
	    
	  case ctlSelectEvent:
	    i = FrmGetObjectIndex(f, FileFileList);
	    if (i != frmInvalidObjectId)
	      l=(ListType *) FrmGetObjectPtr(f, i);
            
	    switch (e->data.ctlSelect.controlID) 
	    {
	      case FileLoadButton:
	      case FileLoadRunButton:
	        if (l) 
	        {
      			i = LstGetSelection(l);
      			if (i>=0)
      			  LoadSelectedFile(i);
      		}
      		doExit = true;
      		if(e->data.ctlSelect.controlID == FileLoadRunButton)
      		  StrCopy(strSendKeysBuffer, "LOAD\":*\",8,1\nRUN\n");
      		break;

	      case FileSaveButton:
      		s = ShowSaveAsDialog(NULL);
      		if (s) 
      		{
      			// When the file dialog is about to go up, the C=64 is alerted
      			// to save a snapshot, then we delay a frame.
      			// Thanks to this armletData->snapshot is valid here.
      			if(lastCard == 0xffff)
      			  SaveSnapshot(s, -1);
      			else
      			  SaveSnapshot2SD(s, lastCard);
        		ScanFiles(f);
      		}
      		break;
      		
	      case FileDeleteButton:
	        // Delete by index (removes all subsequent parts up to the next
      		// valid "first part")
	        if (l) 
	        {
      			i = LstGetSelection(l);
      			DeleteFile(i);
    				ScanFiles(f);
      		}
      		break;
      		
	      case FileExitButton:
	        doExit = true;
      		break;
	    }
	  default:
	    handled = false;
	}
	
	if (doExit) 
	{
    SetKeyMask(); // Back to normal.
		if (l) 
    	listLastIndex = LstGetSelection(l);
		FrmReturnToForm(MainForm);
		SetPaused(false);
	}

	return handled;
}


//***********************************************************
// Procedure: HandleEventGfxSound
//
// Parameter: e   Event
//
// Returnval: True if event handled in this procedure
//
// Description: Handle event of dialog to set graphics and
//              sound preferences.
//***********************************************************
Boolean HandleEventGfxSound(EventPtr e)
{
	UInt32 v;
	FormPtr f = FrmGetActiveForm();
	char msg[16];
	
	switch (e->eType) 
	{
	  case ctlRepeatEvent:
	    switch(e->data.ctlRepeat.controlID)
	    {
  	    case GfxSoundSpeedLimitFeedbackSliderControl:
  		    StrPrintF(msg, "Speed: %d%%", e->data.ctlRepeat.value);
    	    SetText(f, GfxSoundSpeedLimitLabel, msg);
          break;

  	    case GfxSoundVolumeFeedbackSliderControl:
    	    v = e->data.ctlRepeat.value;
    	    if (!v)
        		StrPrintF(msg, "Vol: Mute");
    	    else
    		    StrPrintF(msg, "Vol: %ld%%", (100*v) >> 10);
    	    SetText(f, GfxSoundVolumeLabel, msg);
    	    MakeNoise(true, v);
          break;
      }
	}
	return false;
}


//***********************************************************
// Procedure: HandleEventJoystick
//
// Parameter: e   Event
//
// Returnval: True if event handled in this procedure
//
// Description: Handle event of dialog to set joystick preferences
//***********************************************************
Boolean HandleEventJoystick(EventPtr e)
{
	Boolean handled = false;
	FormType *f = FrmGetActiveForm();

	switch (e->eType) 
	{
	  case ctlSelectEvent:
	    switch(e->data.ctlSelect.controlID) 
	    {
	      case JoystickUpButton:
      		dlgUp = SampleKey(dlgUp);
      		SetText(f, JoystickUpLabel, StringForKey(dlgUp));
      		handled = true;
      		break;
	      case JoystickLeftButton:
      		dlgLeft = SampleKey(dlgLeft);
      		SetText(f, JoystickLeftLabel, StringForKey(dlgLeft));
      		handled = true;
      		break;
	      case JoystickRightButton:
      		dlgRight = SampleKey(dlgRight);
      		SetText(f, JoystickRightLabel, StringForKey(dlgRight));
      		handled = true;
      		break; 
	      case JoystickDownButton:
      		dlgDown = SampleKey(dlgDown);
      		SetText(f, JoystickDownLabel, StringForKey(dlgDown));
      		handled = true;
      		break;
	      case JoystickFireButton:
      		dlgFire = SampleKey(dlgFire);
      		SetText(f, JoystickFireLabel, StringForKey(dlgFire));
      		handled = true;
      		break;
	      case JoystickOKButton:
      		prefsPalm->KeyUp = dlgUp;
      		prefsPalm->KeyLeft = dlgLeft;
      		prefsPalm->KeyRight = dlgRight;
      		prefsPalm->KeyDown = dlgDown;
      		prefsPalm->KeyFire = dlgFire;
      		prefs->Joystick1On = GetValue(f, JoystickEmulate1Checkbox);
      		prefs->Joystick2On = GetValue(f, JoystickEmulate2Checkbox);
      		prefsPalm->JoyPen = (GetValue(f, JoystickJoyPenScreenCheckbox)? JOYPEN_SCREEN:0) |
		        (GetValue(f, JoystickJoyPenGraffitiCheckbox)? JOYPEN_GRAFFITI:0);
      		SavePrefs();
  			  SavePrefsPalm();
      		SetKeyMask();
      		armletData->cmd |= CMD68K_SETPREFS;
      		break;
	    }
	}
	return handled;
}


//***********************************************************
// Procedure: HandleEventKeyboard
//
// Parameter: e   Event
//
// Returnval: True if event handled in this procedure
//
// Description: Handle event of dialog to set keyboard preferences
//***********************************************************
Boolean HandleEventKeyboard(EventPtr e)
{
	Int16 i;
	UInt32 lIndex;
	UInt32 lKey;
	FormType *f = FrmGetActiveForm();

	switch (e->eType) 
	{
	  case ctlSelectEvent:
	    switch(e->data.ctlSelect.controlID) 
	    {
	      case KeyboardOKButton:
      	  SavePrefsPalm();
      		SetKeyMask();
      		break;
	    }
	    break;

	  case penDownEvent:
      if(e->screenY > 14 && e->screenY < 130)
      {
  	    for(i=0; i<4; ++i)
  	    {
  	      if(e->screenX > g_lColPos[2 * i] && e->screenX < g_lColPos[2 * (i + 1)])
  	      {
  	        lIndex = i * 17 + (e->screenY - 14) / 7;
  	        if(lIndex >= 0 && lIndex < 66)
  	        {
    	        for(lKey = 0; lKey < 66; ++lKey)
    	        {
    	          if(C64KeyToDisplay[lKey] == lIndex)
    	            break;
    	        }
    	        prefsPalm->KeyMapping[lKey] = SampleKey(prefsPalm->KeyMapping[lKey]);
    	        FrmDrawForm(f);
            }
  	        break;
  	      }
  	    }
      }
      break;	  
	}

	return false;
}


//***********************************************************
// Procedure: HandleEventSampleKey
//
// Parameter: e   Event
//
// Returnval: Flag if event handled in this procedure
//
// Description: Handle event of Sample-Key-Dialog
//***********************************************************
// Klunky like hell... A requester and three buttons to get a key? Well, it was the easiest way.
Boolean HandleEventSampleKey(EventPtr e)
{
	Boolean handled = false;
	FormType *f = FrmGetActiveForm();
	
	switch (e->eType) 
	{
	  case ctlSelectEvent:
	    switch(e->data.ctlSelect.controlID) 
	    {
	      case SampleKeySampleButton:
      		sampledKey = KeyCurrentState();
		      SetText(f, SampleKeyKeyLabel, StringForKey(sampledKey));
      		handled = true;
		      break;
	    }
	    break;
	}
	return handled;
}


//*******************************************************************************************
//*******************************************************************************************
// Graphics
//*******************************************************************************************
//*******************************************************************************************

//***********************************************************
// Procedure: ChangeScreenMode
//
// Parameter: sm        New screen mode
//            first     Flag if this is the first call
//            defaults
//
// Returnval: 1 if an error occured, 0 if successful
//
// Description: Set new screen mode.
//***********************************************************
UInt32 ChangeScreenMode(UInt32 sm, Boolean first, Boolean defaults)
{
	UInt32 w, h, d = 8;
	Boolean col = true;
	UInt32 defaultD;
	UInt32 oldsm;
	Boolean buffer = (sm == SCREENMODE_256_BUFFERED);
	UInt32 error;
	UInt16 i16;
	
	oldsm = prefsPalm->ScreenMode;
	// Check for unreasonable screen modes
	if (!is320 || !can256) 
	{
		sm = SCREENMODE_256_BUFFERED;
		buffer = true;
	}
	
	if (sm == prefsPalm->ScreenMode && !first && !defaults)
		return 0; // Nothing to change
	
	if (buffer) 
	{
		// Draw to a bitmap
		// Use the default screen mode in this case
		if (!IsNuked) 
		{
			WinScreenMode(winScreenModeSetToDefaults, NULL, NULL, NULL, NULL);
  		WinScreenMode(winScreenModeGetDefaults, NULL, NULL, &defaultD, NULL);
      if(defaultD > 8)
  			WinScreenMode(winScreenModeSet, NULL, NULL, &d, NULL);
		}

		// Wrong size or no 256-color mode. Render to bitmap.
    error=WinPalette(winPaletteSet, 0 /* i*16 */, 16, (RGBColorType *) prefsPalm->colorTable);
    if (error) 
    {
      ShowErrorResLong(StringErrFailedToChangePalette, error);
			return error;
	  }

	  bmp = BmpCreate(320, 272, 8, NULL, &i16);
		if (!bmp) 
		{
			ShowErrorResLong(StringErrBmpCreate, i16);
			return 1;
		}

		bmpv3=BmpCreateBitmapV3(bmp, kDensityDouble, BmpGetBits(bmp), NULL);
		if (!bmpv3) 
		{
			ShowErrorRes(StringErrNoHiresBitmap);
			return 1;
		}
		bmpBits = BmpGetBits((BitmapType *) bmpv3);
		
		if (((UInt32) bmpBits) & 3) 
		{
			ShowErrorRes(StringErrBmpBitsNotAligned);
			return 1;
		}
		if (!bmpBits) 
		{
		  ShowErrorRes(StringErrFailedToGetBmpBits);
			return 1;
		}
		armletData->displayBuf = (UInt8 *) swap32((UInt32) bmpBits);
		armletData->displayModulo = swap32(320);
	} 
	else  
	{
		// Direct memory access
		// If we crashed, or this is the first time Frodo is run
		// don't do any screen mode changes until we get the default prefs
		if (!IsNuked) 
		{
			w = screenW;
			h = screenH;
		  error=WinScreenMode(winScreenModeSet, &w, &h, &d, &col);
		  if (error) 
		  {
    		ShowErrorResLong(StringErrFailedToChangeScreenMode, error);
    		return error;
		  }
		}
		error=WinPalette(winPaletteSet, 0, 16, (RGBColorType *) prefsPalm->colorTable);
		if (error) 
		{
			ShowErrorResLong(StringErrFailedToChangePalette, error);
			return error;
		}
		screenBits = (UInt8 *) BmpGetBits(WinGetBitmap(WinGetDrawWindow()));
		if (!screenBits) 
		{
			ShowErrorRes(StringErrFailedToGetBmpBits);
			return 1;
		}
		if (bmpBits) 
		{
			BmpDelete(bmp);
			BmpDelete((BitmapPtr) bmpv3);
			bmp = NULL;
			bmpv3 = NULL;
			bmpBits = NULL;
		}
		armletData->displayBuf = screenBits;
		armletData->displayModulo = swap32(screenW);
			
		// There's an off chance that the device will report double
		// density but it does in fact have a width other than 320 pixels.
		// <320 was tackled in CheckPrerequisites, >320 right here.
		if (screenW > 320) 
		{
			// ... and it doesn't work at all on eg. UX50 480x320
			// because it LIE's about the true width.
			// With no way to get the true width except device ID.
			// Let them use "compatible mode".
			UInt16 offs = (screenW - 320) / 2;
			UInt16 scroll;
			scroll = offs & 0xfffc; // Center, keep it aligned
			armletData->displayBuf += scroll;
		}
		armletData->displayBuf = (UInt8 *) swap32((UInt32) armletData->displayBuf);
	}
	prefsPalm->ScreenMode = sm;
	if (crashed || !first)
	{
	  SavePrefs();
	  SavePrefsPalm();
	}

	return 0;
}


//***********************************************************
// Procedure: InitGraphicsSound
//
// Parameter: -
//
// Returnval: Errorcode of 0 if successful
//
// Description: Initialize graphics and sound
//***********************************************************
UInt32 InitGraphicsSound()
{
	UInt32 error;
	UInt16 i;
		
	error = ChangeScreenMode(prefsPalm->ScreenMode, true, false);
	if (error)
		return error;

	// Preload joystick bitmaps
	for (i=0; i<10; i++)
	{
		joyHdl[i] = GetResource(bitmapType, i+32);
		if (joyHdl[i])
			joyPtr[i] = MemHandleLock(joyHdl[i]);
  }
    
  // Load standard font from resources
  hFontStandard = GetResource(fontExtRscType, fntStandard);
  if(hFontStandard)
  {
    pFontStandard = (FontType *) MemHandleLock(hFontStandard);
    if(pFontStandard)
    {
      FntDefineFont(fntStandardID, pFontStandard);
	    lOldFont = FntSetFont(fntStandardID);
	  }
	}

	// Now the sound
	return InitSound();
}


//***********************************************************
// Procedure: FreeGraphicsSound
//
// Parameter: -
//
// Returnval: -
//
// Description: Shut of audio, release bitmaps, reset screen mode
//***********************************************************
void FreeGraphicsSound(void)
{
	UInt16 i;

	// Release bitmaps to dislay joystick
	for (i=0; i<10; i++) 
	{
		if (joyHdl[i])
		{
			if (joyPtr[i])
				MemHandleUnlock(joyHdl[i]);
			FreeResource(joyHdl[i]);
		}
	}

  // Free resources of fonts
	if(0 != lOldFont)
	{
		// Reset old font
		FntSetFont(lOldFont);
		lOldFont = 0;
	}
	if(pFontStandard)
	  MemHandleUnlock(hFontStandard);
	pFontStandard = NULL;
  if(hFontStandard)
  	FreeResource(hFontStandard);
	hFontStandard = NULL;

	WinScreenMode(winScreenModeSetToDefaults, NULL, NULL, NULL, NULL);
	if (bmpBits) 
	{
		BmpDelete(bmp);
		BmpDelete((BitmapPtr) bmpv3);
		bmp = NULL;
		bmpv3 = NULL;
		bmpBits = NULL;
	}
	
	FreeSound();
}


//***********************************************************
// Procedure: RepaintStatistics
//
// Parameter: speed percentage
//
// Description: Draw statistics. Currently only performance.
//***********************************************************
void RepaintStatistics(UInt32 percentage)
{
	static RectangleType statRect = {{133,136},{19,6}};
	char msg[12];
	
	WinSetBackColor(0); 	// Black
	WinSetForeColor(12);  // Grey
	WinSetTextColor(12);  // Grey
	WinEraseRectangle(&statRect, 0);
	if (prefsPalm->paused)
	{
		WinDrawChars("----", 6, 134, 138);
	} 
	else 
	{
		if(percentage > 999)
		  percentage = 999;
		StrPrintF(msg, "%3ld%%", percentage);
		WinDrawChars(msg,StrLen(msg), 134, 137);
	}
}


//***********************************************************
// Procedure: RepaintTrackNum
//
// Parameter: Current track number and LED state
//
// Description: Draw track number.
//***********************************************************
void RepaintTrackNum(UInt32 track, UInt8 led)
{
	static RectangleType statRect = {{133,136},{19,6}};
	char msg[12];

	WinSetBackColor(0); 	// Black
	WinSetForeColor(12); 	// Grey
	if(led)
	  WinSetTextColor(5); // Green
	else
	  WinSetTextColor(12); 	// Grey
	WinEraseRectangle(&statRect, 0);
	StrPrintF(msg, "Tr%2ld%", track);
	WinDrawChars(msg,StrLen(msg), 134, 137);
}


//***********************************************************
// Procedure: RepaintSCIndicator
//
// Parameter: SCOn  Flag if single cycle emulation is on
//
// Returnval: -
//
// Description: Display indicator for single cycle on screen. 
//***********************************************************
void RepaintSCIndicator(Boolean SCOn)
{
  MemHandle hdl = 0;
  BitmapPtr ptr;
	FormPtr f;
  DmResID ind;
  
  if(SCOn)
    ind = IndicatorSCBitmapFamily;
  else
    ind = IndicatorNoSCBitmapFamily;

	hdl = GetResource(bitmapType, ind);
	if (hdl)
	{
		ptr = MemHandleLock(hdl);
		if(ptr)
    {
    	f = FrmGetActiveForm();
      if (FrmGetFormId(f) == MainForm)
      {
        WinDrawBitmap(ptr, 151, 136);
      }
      MemHandleUnlock(hdl);
    }
    FreeResource(hdl);  
  }
}


//***********************************************************
// Procedure: Repaint1541Indicator
//
// Parameter: On1541  Flag if 1541 emulation is on
//
// Returnval: -
//
// Description: Display indicator for 1541 on screen. 
//***********************************************************
void Repaint1541Indicator(Boolean On1541)
{
  MemHandle hdl = 0;
  BitmapPtr ptr;
	FormPtr f;
  DmResID ind;
  
  if(On1541)
    ind = Indicator1541BitmapFamily;
  else
    ind = IndicatorNo1541BitmapFamily;

	hdl = GetResource(bitmapType, ind);
	if (hdl)
	{
		ptr = MemHandleLock(hdl);
		if(ptr)
    {
    	f = FrmGetActiveForm();
      if (FrmGetFormId(f) == MainForm)
      {
        WinDrawBitmap(ptr, 151, 139);
      }
      MemHandleUnlock(hdl);
    }
    FreeResource(hdl);  
  }
}


//***********************************************************
// Procedure: RepaintOnScreenKeyboard
//
// Parameter: f   Form to display keyboard
//
// Returnval: -
//
// Description: Display keyboard on screen. 
//***********************************************************
void RepaintOnScreenKeyboard(FormType *f)
{
	RectangleType r;
	UInt16 idx = FrmGetObjectIndex(f, keyboardBmp);
	MemHandle m = 0;
	BitmapType *bmp;
	
	if (idx==frmInvalidObjectId)
		return;
	
	FrmGetObjectBounds(f, idx, &r);
	m = GetResource(bitmapType, keyboardBmp);
	if (m) 
	{
		bmp = MemHandleLock(m);
		if (bmp) 
		{
			WinDrawBitmap(bmp, r.topLeft.x, r.topLeft.y);
			MemHandleUnlock(m);
		} 
		else
			ShowErrorRes(StringErrFailedToLockHandle);
		FreeResource(m);
	}
	else
		ShowErrorResStr(StringErrResourceNotFound, "Tbmp");

	return;
}


//***********************************************************
// Procedure: UpdateOnScreenKeyboard
//
// Parameter: f   Form to display keyboard
//
// Returnval: -
//
// Description: Sets the correct keyboard underlay. 
//***********************************************************
void UpdateOnScreenKeyboard(FormType *f)
{
	UInt16 idx = FrmGetObjectIndex(f, keyboardBmp);

	if (idx==frmInvalidObjectId)
		return;
  FrmRemoveObject(&f, idx);
	
	if (cmdLock) 
		keyboardBmp = KeyboardCmdBitmap;
	else if (lowerCase) 
	{
		if (shiftLock) 
			keyboardBmp = KeyboardLocaseShiftBitmap;
		else
			keyboardBmp = KeyboardLocaseBitmap;
	} 
	else if (shiftLock) 
		keyboardBmp = KeyboardShiftBitmap;
	else 
		keyboardBmp = KeyboardBitmap;

	if (!FrmNewBitmap(&f, keyboardBmp, keyboardBmp, 0, 136))
		return;

	RepaintOnScreenKeyboard(f);
}


//***********************************************************
// Procedure: UpdateOnScreenJoystick
//
// Parameter: -
//
// Returnval: -
//
// Description: Display joystick state on screen. 
//***********************************************************
void UpdateOnScreenJoystick(void)
{
	UInt32 joy = swap32(armletData->joystick);
	FormPtr f;

	if (joy == oldJoy) 
		return;
	oldJoy = joy;
	
	f = FrmGetActiveForm();
	if (FrmGetFormId(f) != MainForm)
	  return;

	if (joy & JOYUP) 
	{
		if (joy & JOYLEFT)
			joyBmp = 40;
		else if (joy & JOYRIGHT)
			joyBmp = 34;
		else
			joyBmp = 33;
	} 
	else if (joy & JOYDOWN) 
	{
		if (joy & JOYLEFT)
			joyBmp = 38;
		else if (joy & JOYRIGHT)
			joyBmp = 36;
		else
			joyBmp = 37;
	} 
	else 
	{
		if (joy & JOYLEFT)
			joyBmp = 39;
		else if (joy & JOYRIGHT)
			joyBmp = 35;
		else
			joyBmp = 32;
	}
	if (joyPtr[joyBmp-32])
	{
	  if(prefs->Joystick1On)
	  {
		  WinDrawBitmap(joyPtr[joyBmp-32], 135, 143);
		  if(joy & JOYFIRE)
  		  WinDrawBitmap(joyPtr[9], 138, 146);
		}
		else
		{
		  WinDrawBitmap(joyPtr[joyBmp-32], 147, 143);
		  if(joy & JOYFIRE)
  		  WinDrawBitmap(joyPtr[9], 150, 146);
		}
	}
}


//*******************************************************************************************
//*******************************************************************************************
// Sound
//*******************************************************************************************
//*******************************************************************************************

//***********************************************************
// Procedure: InitSound
//
// Parameter: -
//
// Returnval: Errorcode or 0 if successful
//
// Description: Initialize sound system.
//***********************************************************
UInt32 InitSound(void)
{
	// Set the buffer even if we won't use it (just in case...)
	armletData->soundBuf = NULL;
	if (!(soundStruct.soundBuf = MemPtrNew(SOUND_BUFSIZE*2))) 
	{
		ShowErrorRes(StringErrNoSoundBuffer);
		canSound=false;
	}
	else
		armletData->soundBuf = (UInt16 *) swap32((UInt32) soundStruct.soundBuf);
	soundStruct.nextWritePos = soundStruct.soundBuf;
	
	if (canSound)
		StartSound();
	
	return 0;
}


//***********************************************************
// Procedure: FreeSound
//
// Parameter: -
//
// Returnval: -
//
// Description: Release sound system.
//***********************************************************
void FreeSound(void)
{
	if (sndStream) 
		SndStreamDelete(sndStream);
	if (soundStruct.soundBuf)
		MemPtrFree(soundStruct.soundBuf);
}


//***********************************************************
// Procedure: StartSound
//
// Parameter: -
//
// Returnval: -
//
// Description: Start playback of sound.
//***********************************************************
void StartSound(void)
{
	Err error;
	
	if (sndStream) 
	{
		// Stream exists. Assume it was paused.
		SndStreamPause(sndStream, false);
		SndStreamSetVolume(sndStream, prefsPalm->Volume);
		return;
	}

	// SoundCallback will be called whenever the hw needs samples.
	// We have no way of knowing how many samples will be demanded,
	// and must deliver them STAT. Oh joy.
	error = SndStreamCreate(&sndStream, sndOutput, 44100, sndInt16Little, sndMono, SoundCallback, &soundStruct,
			prefsPalm->SoundBufferSize, false);
	if (error) 
	{
		ShowErrorRes(StringErrFailedToStartSound);
		sndStream=NULL;
	} 
	else 
	{
		SndStreamStart(sndStream);
		SndStreamSetVolume(sndStream, prefsPalm->Volume);
	}
}


//***********************************************************
// Procedure: StopSound
//
// Parameter: -
//
// Returnval: -
//
// Description: Stop playback of sound.
//***********************************************************
void StopSound(void)
{
	if (sndStream)
		SndStreamPause(sndStream, true);
}


//***********************************************************
// Procedure: PauseSound
//
// Parameter: b   Flag if playback stopped or resumed
//
// Returnval: -
//
// Description: Pause/resume playback of sound.
//***********************************************************
void PauseSound(Boolean b)
{
	if (b)
		StopSound();
	else if (sndStream)
		StartSound();
}


//***********************************************************
// Procedure: MakeNoise
//
// Parameter: b   Flag for make noise or stop noise
//            v   Current volume
//
// Returnval: -
//
// Description: Make noise to help user adjusting sound volume.
//***********************************************************
void MakeNoise(Boolean b, UInt32 v)
{
  soundStruct.makeNoise = b;
  if(b)
    SndStreamSetVolume(sndStream, v);
}


//***********************************************************
// Procedure: SoundCallback
//
// Parameter: -
//
// Returnval: Errorcode or 0 if successful
//
// Description: Callback from PalmOS for next sound samples.
//***********************************************************
/* Explanation: (See also VBlank) the SoundStruct contains a large(ish)
   ringbuffer of samples. SampleOffset is the index of the next sample that
   hasn't been output. SamplesQueued inidicates how many samples have been
   produced by the armlet beyond that point.
   SamplesQueued can never grow larger than the size of the buffer
   (if it does the data is thrown away rather than being queued.)
   The parameter 'size' is how many samples we must write to bufferP.
   Since I just fill any underflow with zeros, the net effect of wildly
   fluctuating frame rates should be that the latency of the stream 
   automatically goes up (amidst horrible clicking).
   After a while the audio should be nice again (but high latency).
   This will ONLY work if the buffer size is larger than the maximum lag!
   Currently 4kB=0.1 s, kinda low.
   In case of pause, modal dialogs or the like the ringbuffer will quickly be
   emptied, and the whole process starts over when the emulator starts running
   normally again.
   If the Palm simply isn't fast enough to run at an average speed of 50Hz
   (even when dropping 90% of the frames) the sound WILL be shot to hell.
   That's what the 'Mute' option is for.
*/
Err SoundCallback(void *userDataP, SndStreamRef stream, void *bufferP, UInt32 size)
{
	struct SoundStruct *ss = (struct SoundStruct *) userDataP;
	UInt16 *p = (UInt16 *)bufferP;
	UInt32 nsmpl = size, nempty = 0;
	UInt32 i, j;

	// Feedback noise when adjusting volume
	if (ss->makeNoise) 
	{
		// ~800 Hz square wave, 64 samples per wave
		i = 0;
		// Finish previous half-cycle
		j = 32-(ss->noiseOffs&31);
		if (j>size) 
			j=size;
		if (ss->noiseOffs&32)
			MemSet(p, j*2, 0x1f);
		else
			MemSet(p, j*2, 0xe0);
		i += j;
		ss->noiseOffs = (ss->noiseOffs+j)&63;
		
		// Produce square waves, the last will be truncated.
		while (i<size) 
		{
			j = 32;
			if (i+j > size) 
				j = size-i;
			// 1/4 max volume, like a single SID voice.
			if (ss->noiseOffs&32)
				MemSet(p+i, j*2, 0x1f);
			else
				MemSet(p+i, j*2, 0xe0);
			i += j;
			ss->noiseOffs = (ss->noiseOffs+j)&63;
		}
		return errNone;
	}
	
	// The real deal
	if (size > ss->samplesQueued) 
	{
		// Underflow, fill with 0 
		nempty = size-ss->samplesQueued;
		nsmpl = ss->samplesQueued;
	}
	if (ss->sampleOffset+nsmpl > SOUND_BUFSIZE) 
	{
		MemMove(p, ss->soundBuf+ss->sampleOffset, (SOUND_BUFSIZE - ss->sampleOffset)*2);
		MemMove(p+SOUND_BUFSIZE - ss->sampleOffset, ss->soundBuf, (ss->sampleOffset+nsmpl - SOUND_BUFSIZE)*2);
		ss->sampleOffset += nsmpl;
		ss->sampleOffset -= SOUND_BUFSIZE;
	} 
	else 
	{
		MemMove(p, ss->soundBuf+ss->sampleOffset, nsmpl*2);
		ss->sampleOffset += nsmpl;
	}
	// It seems an idle SID produces 0xe000
	MemSet(p+nsmpl, nempty*2, 0xdf);
	
	if (size > ss->samplesQueued)
	{
		ss->sampleOffset = ss->nextWritePos - ss->soundBuf;
	}
	ss->samplesQueued -= nsmpl;

	return errNone;
}


//***********************************************************
// Procedure: SoundCurrFrame
//
// Parameter: -
//
// Returnval: -
//
// Description: Adjust sound for current frame.
//***********************************************************
void SoundCurrFrame(void)
{
	if (canSound && prefsPalm->Volume && soundStruct.samplesQueued + FRAME_SAMPLES <= SOUND_BUFSIZE) 
	{
		// Explanation:
		//   soundBuf is a large(ish) circular buffer that's filled here
		//   with data from the armlet and drained by the sound callback.
		//   SampleOffset is the offset (in samples) to the next sample
		//   that will be played. SamplesQueued is the number of samples
		//   beyond this point that are ready for consumption.
		soundStruct.nextWritePos = (UInt16 *) swap32((UInt32) armletData->soundBuf);
		soundStruct.nextWritePos = soundStruct.nextWritePos + FRAME_SAMPLES;
		if(soundStruct.nextWritePos - soundStruct.soundBuf >= SOUND_BUFSIZE)
			soundStruct.nextWritePos = soundStruct.soundBuf;
		armletData->soundBuf = (UInt16 *) swap32((UInt32) soundStruct.nextWritePos);

		soundStruct.samplesQueued += FRAME_SAMPLES;
	}
}


//*******************************************************************************************
//*******************************************************************************************
// Callbacks from Armlet
//*******************************************************************************************
//*******************************************************************************************

//***********************************************************
// Procedure: ShowRequester68k
//
// Parameter: p   Array with information to display
//
// Returnval: Always 0
//
// Description: Display one of the requesters. Just ugly FatalAlerts right now.
//***********************************************************
Int32 ShowRequester68k(Int32 *p)
{
	Int32 *ip, id, arg1, arg2, arg3, arg4;
	char buf[11];
	int i;
	char msg[128];
	
	if (!p) 
	{
		StrPrintF(msg, "Illegal call to ShowRequester, parameter is NULL");
		ShowError(msg);
		return 0;
	}
	ip = (Int32 *) swap32((UInt32) p);
	id = swap32(ip[0]);
	arg1 = swap32(ip[1]); 
	arg2 = swap32(ip[2]);
	arg3 = swap32(ip[3]); 
	arg4 = swap32(ip[4]);
	
	switch (id) 
	{
		case REQ_BAD_SNAPSHOT:
			StrPrintF(msg, "Not a valid Frodo snapshot (\"%s\").", arg1);
			break;
		case REQ_BAD_SNAPSHOT_VERSION:
			StrPrintF(msg, "Invalid snapshot version %ld.", arg1);
			break;
		case REQ_BAD_OPCODE:
			StrPrintF(msg, "Illegal opcode 0x%04x at 0x%04x.", ((UInt16)arg1), ((uint16)arg2));
			break;
		case REQ_BAD_JUMP:
			StrPrintF(msg, "Jump to I/O space at 0x%04x to 0x%04x.", ((UInt16)arg1), ((uint16)arg2));
			break;
		case REQ_BAD_OPCODE_1541:
			StrPrintF(msg, "1541: Illegal opcode 0x%04x at 0x%04x.", ((UInt16)arg1), ((uint16)arg2));
			break;
		case REQ_BAD_JUMP_1541:
			StrPrintF(msg, "1541: Jump to I/O space at 0x%04x to 0x%04x.", ((UInt16)arg1), ((uint16)arg2));
			break;
		case REQ_NEW_DRIVE:
			StrPrintF(msg, "Out of memory changing disk %d.",	((UInt16)arg1));
			break;
		case REQ_NEW_SIDRENDERER:
			StrPrintF(msg, "Out of memory when starting audio.");
			break;

		// DEBUG requesters below this point
		case REQ_DEBUG:
			StrPrintF(msg, "Debug: %lx %lx %lx %lx", arg1, arg2, arg3, arg4);
			break;
		case REQ_DEBUGSTR:
			StrPrintF(msg, "Debug \"%s\" %lx %lx %lx", arg1, arg2, arg3, arg4);
			break;
		case REQ_EL_TEXT:
			for (i=0; i<10; i++)
				buf[i]=((char *)arg2)[i];
			buf[10]='\0';
			StrPrintF(msg, "EL_*_TEXT: %lx (screen=%lx) <- %lx (\"%s\"), shift=%ld", arg1, screenBits, arg2, buf, arg3);
			break;
		case REQ_EL_GFX:
			for (i=0; i<10; i++)
				buf[i]=((char *)arg2)[i];
			buf[10]='\0';
			StrPrintF(msg, "EL_*: %lx (screen=%lx) <- %lx, shift=%ld", arg1, screenBits, arg2, arg3);
			break;
		default:
			StrPrintF(msg, "Unknow req %ld (%ld,%ld,%ld,%ld)", arg1, arg2, arg3, arg4);
	}
	// TODO 1541 errors should have two buttons "Reset C64", "Reset 1541"
	// I really should draw real Alrt's for these...
	ShowError(msg);

	return 0;
}


//***********************************************************
// Procedure: Malloc68k
//
// Parameter: n   Size of memory
//
// Returnval: Pointer to allocated memory
//
// Description: Allocate memory for armlet side
//***********************************************************
void *Malloc68k(UInt32 n)
{
	void *p = MemPtrNew(swap32(n));
	if(!p)
	  ShowErrorRes(StringErrNoMemForARM);
	return p;
}


//***********************************************************
// Procedure: Free68k
//
// Parameter: Pointer to memory to free
//
// Returnval: Always 0
//
// Description: Free memory for armlet side
//***********************************************************
UInt32 Free68k(void *p)
{
  if(p != NULL)
	  MemPtrFree((void *) swap32((UInt32)p));
	return 0;
}


//***********************************************************
// Procedure: DmWrite68k
//
// Parameter: p   Pointer to array with parameter for saving
//
// Returnval: Errorcode of 0 if successful
//
// Description: Write data back to PalmDatabase.
//***********************************************************
Int32 DmWrite68k(Int32 *p)
{
  Int32 *ip;
  uint8 *pDest;
  UInt32 lOffset;
  uint8 *pSrc;
  UInt32 lSize;
  UInt32 err = 0;
  
  // Convert parameter from armlet to 68K
  ip = (Int32 *)swap32((UInt32)p);
  pDest = (uint8 *) swap32(ip[0]);
  lOffset = (UInt32) swap32(ip[1]);
  pSrc = (uint8 *) swap32(ip[2]);
  lSize = (UInt32) swap32(ip[3]);
  
  // Write data to record (52: Header in record of PalmDatabase)
  err = DmWrite(pDest - 52, lOffset + 52, pSrc, lSize);

  // Return status to armlet
  return err;
}


//***********************************************************
// Procedure: VBlank68k
//
// Parameter: p   Pointer to armlet data (armlet endian!)
//
// Returnval: Always 0
//
// Description: Called every frame by the armlet. Our only chance to handle I/O.
//***********************************************************
UInt32 VBlank68k(void *p)
{
	static UInt32 frame = 0, tps, tpf, tpf100;
	static UInt32 startTick = 0;
	static UInt32 lastTick = 0;
	static UInt32 didDraw = 0, didSkip = 0;
	static UInt32 perfTick = 0, perfStat = 0;
	static UInt8 nskipped = 0;
	static Int16 speedLimit = 100;
	Int32 diff;
	UInt32 tick, sleepTo = 0;
	char *s;
	UInt32 i;
	EventType e;
	UInt16 menuError;
	UInt32 lTicksOnEnter = TimGetTicks() * 10;
	UInt8 key;
	
	// Initialization at frame==0
	if (frame == 0L) 
	{
		tps = SysTicksPerSecond() * 10;
		tpf = (tps * 100) / (50 * speedLimit);
		tpf100 = tps/50; // ticks per frame for 100% of C64 speed
		// Timing mismatch if tps isn't a multiple of fps
		diff = 100*tps - tpf*5000;
		if (diff > 3 || diff < -3) // >3% off optimal
			ShowErrorResLong(StringWarnBadTimingParam, tps);
		startTick = TimGetTicks() * 10;
	}

  if(speedLimit != prefsPalm->SpeedLimit)
  {
    // New speed limit -> recalc
    speedLimit = prefsPalm->SpeedLimit;
    tpf = (tps * 100) / (50 * speedLimit); 
  }
  
	// restore paused state from preferences after first frame is drawn
	if (!pausedStateValid && armletData->cmd&CMD68K_DRAWFRAME) 
	{
		SetPaused((prefsPalm->paused != 0));
		pausedStateValid = true;
	}

	// Show the about dialog first time, after the C=64 has initialized itself
	if(firstTime && frame == 100)
		ShowDialog(AboutForm);

	// I do this first to get maximum precision in the clocks.
	// sleepTo is when this frame should end, in a perfect world.
	// It's lastTick+1/50 s. If we lag very badly (10 frames right now),
	// the timing is reset.
	if (!lastTick || nskipped >= 20) 
	{
		// First time, or after pause/dialog, or just a slow cpu.
		lastTick = tick = sleepTo = TimGetTicks() * 10;
		nskipped = 0;
	} 
	else 
	{
		tick = TimGetTicks() * 10;
		sleepTo = lastTick + tpf;
	}
	
	if (!prefs->LimitSpeed)
		// Turbo mode, no sleep
		sleepTo = tick;

	// Send hint to armlet to skip next frame if we're running late
	armletData->cmd &= ~(CMD68K_DRAWFRAME);
	if (prefs->AdaptiveFrameSkip) 
	{
		// Adaptive mode, draw next frame if we're on time, otherwise skip.
		if (tick <= sleepTo)
			armletData->cmd |= CMD68K_DRAWFRAME;
	} 
	else 
	{
		// Old-style, use user-specified frame skip ratios.
		if (nskipped+1 >= swap32(prefs->SkipFrames))
			armletData->cmd |= CMD68K_DRAWFRAME;
	}
	
	if (armletData->cmd & CMD68K_GETPREFS) 
	{
		// We're unsure of the prefs (esp. screen mode), so don't draw anything
		// until we get a set of prefs.
		armletData->cmd &= ~CMD68K_DRAWFRAME;
	}
	if (armletData->cmd & CMD68K_DRAWFRAME) 
	{
		nskipped = 0;
		didDraw++;
	} 
	else 
	{
		nskipped++;
		didSkip++;
	}

	// Sound, don't overqueue in turbo mode
  SoundCurrFrame();
	if(soundStruct.samplesQueued < prefsPalm->SoundBufferSize && prefsPalm->Volume)
	{
		sleepTo = tick; // Sound buffer empty -> fill up (no wait)
		armletData->cmd &= ~(CMD68K_DRAWFRAME); // and don't draw next frame
	}
	
	// Debuggish, performance
	if ((frame & 31) == 0) 
	{
		// compute emulation speed
		// Should really display CPU speed when turbo on
		if (!prefs->LimitSpeed) 
		{
			// Unlimited speed. Show actual emulation speed compared to C64
			if(tick - perfTick != 0)
			  perfStat = 100L * 32 * tpf100 / (tick - perfTick);
		} 
		else if (prefs->AdaptiveFrameSkip) 
		{
			// Adaptive frame skip. Show % of frames drawn.
			perfStat = 100 * didDraw / 32;
		} 
		else 
		{
			// Manual frame skip, no turbo. Show... What?
			if(tick - perfTick != 0)
			  perfStat = 100L * 32 * tpf100 / (tick - perfTick);
		}
		perfTick = tick;
		didSkip = didDraw = 0;

		if(prefs->Emul1541Proc && prefsPalm->DisplayTrackNum)
			// draw track number
			RepaintTrackNum(armletData->track1541, armletData->driveLED);
		else
			// draw statistics (speed)
			RepaintStatistics(perfStat);
    
    // Repaint indicators
    RepaintSCIndicator(prefs->SingleCycleEmulation);
    Repaint1541Indicator(prefs->Emul1541Proc);
    
		// reset timer if any key was pressed 
		if (keyStateChanged)
		{
			// reset device auto off timer
			EvtResetAutoOffTimer();
			keyStateChanged = false;
		}
	}
	
	frame++;
	
	// Display
	if ((armletData->cmd & CMD68K_DRAWFRAME) && bmpBits) 
	{
		// Not rendering directly to screen, so blit bitmap
		// We need to handle density better here
		if(screenW >= 320)
			WinDrawBitmap((BitmapPtr)bmpv3, (screenW - 320) / 4, 0);
		else
			WinDrawBitmap((BitmapPtr)bmpv3, (screenW - 160) / 2, 0);
	}

	// Shovel prefs
	if (armletData->cmd & CMD68K_GETPREFS) 
	{
		// We ordered a set of prefs, well, here they are!
		armletData->cmd &= ~(CMD68K_GETPREFS);
		if (prefsPalm->Volume && !canSound) 
			prefsPalm->Volume = 0;

		// Try to get default screen mode based on device id
		prefsPalm->ScreenMode = SCREENMODE_256;
		IsNuked = false;
		ChangeScreenMode(prefsPalm->ScreenMode, false, true);
		if (FrmGetActiveForm()) 
			FrmDrawForm(FrmGetActiveForm());
		SetKeyMask();
		SavePrefsPalm();
	}

	// Snapshot creation (File manager and quicksave)
	if (armletData->cmd & CMD68K_SAVE) 
	{
		// The file manager asked for a snapshot before displaying.
		// Nothing special needs to be done
		armletData->cmd &= ~CMD68K_SAVE;
	}
	
	if(doFileLoadAfterSnap)
	{
    doFileLoadAfterSnap = false;
    if(CheckFileExists(prefs->DriveCard, prefs->DriveIndex, prefs->DrivePath))
      LoadSpecifiedFile(prefs->DriveCard, prefs->DriveIndex, prefs->DrivePath);
    else
    {
  		// Next lines required if an error occured during load
  		FrmGotoForm(MainForm);
  		if (FrmGetActiveForm()) 
  			FrmDrawForm(FrmGetActiveForm());
    }
	}
	
	// The File form delays one frame, so that the armlet can make a snapshot first.
	if (doFileForm) 
	{
    SetPaused(true);
		doFileForm = false;
		MenuEraseStatus(0);
		FrmPopupForm(FileForm);
	}
	
	// Same with save as
	if (doSaveAs) 
	{
		int l = StrLen(lastName);
		int sl, i, n = 0, m = 1;
		char serial[8];
		char snapName[32];
		
		// use temporary name until snapshot is really saved
		StrNCopy(snapName, lastName, 31);
		snapName[31] = 0;
		doSaveAs = false;
		MenuEraseStatus(0);
		
		// We should make up a default name, such as "Game quicksave 1"
		// Clippy style 'AI' follows
	
		// strip .d64/.t64 if present
		if (l >= 4 && snapName[l-4] == '.')
			l -= 4;
		// read serial if present
		for (i = l-1; i >= 0; i--) 
		{
			if (snapName[i] < '0' || snapName[i] > '9')
				break;
			n += m * (snapName[i] - '0');
			m *= 10;
		}
		l = i+1;
		if (n == 0) 
			n=1;
		// append serial+1 to name
		if (l < 31 && (l == 0 || snapName[l-1] != ' ')) // Insert space if appropriate
			StrPrintF(serial, " %d", n+1);
		else
			StrPrintF(serial, "%d", n+1);
		sl = StrLen(serial);
		if (l + sl > 31) // Truncate lastLoad if nessecary
			l = 31 - sl;
		StrCopy(snapName+l, serial); // add serial
			
		s = ShowSaveAsDialog(snapName);
		if (s)
		{
			if(lastCard == 0xffff)
  			// this sets also last used name and index
  			SaveSnapshot(s, -1);
			else
			  SaveSnapshot2SD(s, lastCard);
		}
	}

	if (doQuicksave)
	{
		doQuicksave = false;
		if (lastIndex > 0 || lastCard != 0xffff)
		{
			char strAsk[128];
			
			StrPrintF(strAsk, "Overwrite snapshot '%s' with current state?", lastName);
			i = AskUserDialog(strAsk);
			if (i)
			{
			  // overwrite last loaded/saved position
				if(lastCard == 0xffff)
				  SaveSnapshot(lastName, (UInt16) lastIndex);
   			else
   			  SaveSnapshot2SD(lastName, lastCard);
			}
		}
		MenuEraseStatus(0);
	}
	
	if (doQuickload)
	{
		doQuickload = false;
		if (lastIndex > 0 || lastCard != 0xffff)
		{
			DbFile dbf;
			char strAsk[128];
			
			StrPrintF(strAsk, "Load snapshot '%s' and loose current state?", lastName);
			i = AskUserDialog(strAsk);
			if (i)
			{
				if(lastCard != 0xffff)
				{
				  // load last snapshot from sd
				  StrPrintF(strAsk, "%s.snp", lastName);
				  LoadFile(lastCard, strAsk);
				}
				else
				{
  				// load last loaded/saved position from internal
  				DbFileFromDbRef(mainDBRef, &dbf);
  				Load(&dbf, (UInt16) lastIndex, lastCard);
        }
			}
		}
		MenuEraseStatus(0);
	}

	// No key events
	armletData->key = swap32(0x00ffff);

	// Handle queued up events
	do 
	{
		tick = TimGetTicks() * 10;
		// Mind signed/unsigned 16/32 issues!
		// This simple line caused me a lot of grief :-)
		if (sleepTo > tick + tps) 
		{
			// Sleep more than a second? Something very strange just happened....
			// int32 roll-over? Second coming of Christ?
			sleepTo = tick;
		}
		// Get event, sleepTo-tick is the max we should wait
		if (pause)
			EvtGetEvent(&e, evtWaitForever);
		else if (tick >= sleepTo)
			EvtGetEvent(&e, 0);
		else
			EvtGetEvent(&e, (sleepTo-tick) / 10);

		// return to armlet if no more events queued
		if (e.eType == nilEvent && !pause) 
			break;

		// Standard Palm Event handling
		if (!SysHandleEvent(&e))
			if (!MenuHandleEvent(0, &e, &menuError))
				if (!AppHandleEvent(&e))
					FrmDispatchEvent(&e);

	} while (e.eType != appStopEvent && !g_bSystemOffRequested);
	
	// Joystick
	armletData->joystick = 0;
	CheckJoyPen();
	CheckJoyKey();
	UpdateOnScreenJoystick();
	
	// Mapping Palm-key to C64-key
	CheckKeyKey();

	// Graffiti key-up
	// Time to enqueue penUpEvent? Wait some ticks after KEYDOWN then --> KEYUP
	if(ticksLastKeyDown && (ticksLastKeyDown + 15 > tick / 10) && armletData->key == swap32(0x00ffff)) 
	{
		armletData->key = savedGraffitiKey;
		savedGraffitiKey = 0;
		ticksLastKeyDown = 0;
	}
	
	if(*pSendKeysBufferPos)
	{
	  // Send key to C64 from buffer
	  if(bSendKeysKeyDown == false)
	  {
	    // Send key down to C64
	    key = C64KeyNumberFromAscii[(UInt16) *pSendKeysBufferPos];
	    armletData->key = swap32((KEYDOWN<<16) | ((key_byte[key&0x7f]<<8) | ((key&128)?0x800:0)) | key_bit[key&0x7f]);
	    bSendKeysKeyDown = true;
	  }
	  else
	  {
	    // Send key up to C64
	    key = C64KeyNumberFromAscii[(UInt16) *pSendKeysBufferPos];
	    armletData->key = swap32((KEYUP<<16) | ((key_byte[key&0x7f]<<8) | ((key&128)?0x800:0)) | key_bit[key&0x7f]);
	    bSendKeysKeyDown = false;
	    
	    // Move to next character
	    ++pSendKeysBufferPos;
	    if(*pSendKeysBufferPos == 0)
	    {
	      // End of string reached
	      MemSet(strSendKeysBuffer, SENDKEYSBUFFERSIZE, 0);
	      pSendKeysBufferPos = strSendKeysBuffer;
	    }
	  }
	}
	
	// Normal exit, create auto snapshot
	if (e.eType == appStopEvent || g_bSystemOffRequested) 
	{
		armletData->errCode = swap32(ERR_NORMAL_EXIT);
		armletData->cmd |= CMD68K_SAVE;
	}
	lastTick=sleepTo;
	
	if(TimGetTicks() * 10 - lTicksOnEnter > 10 * tpf)
	{
		// User handled with menu or dialogs
		lastTick = TimGetTicks() * 10 - tpf;
		perfTick = tick;
		didSkip = 0;
		didDraw = 0;
	}
	
	return 0;
}
